/**
 *
 *  Copyright 2004 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.builder;

import java.util.HashMap;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.tranql.cache.CacheSlot;
import org.tranql.cache.CacheTable;
import org.tranql.cache.EndSlot;
import org.tranql.cache.GlobalSchema;
import org.tranql.cache.RelationshipSlot;
import org.tranql.ejb.EJB;
import org.tranql.ejb.EJBSchema;
import org.tranql.ql.QueryException;
import org.tranql.query.CommandTransform;
import org.tranql.query.SchemaMapper;
import org.tranql.query.UpdateCommand;
import org.tranql.schema.Association;
import org.tranql.schema.AssociationEnd;
import org.tranql.schema.Attribute;
import org.tranql.schema.Entity;
import org.tranql.sql.SQLSchema;
import org.tranql.sql.Table;
import org.tranql.sql.UpdateCommandBuilder;


/**
 * Helper class to load a GlobalSchema from an EJBSchema and SQLSchema.
 *
 * @version $Revision: 323 $ $Date: 2006-02-19 01:29:24 -0500 (Sun, 19 Feb 2006) $
 */
public class GlobalSchemaBuilder {
    private final GlobalSchema globalSchema;
    private final EJBSchema ejbSchema;
    private final SQLSchema sqlSchema;
    private final EJBQueryBuilder queryBuilder;
    private final CommandTransform mapper;
    
    public GlobalSchemaBuilder(GlobalSchema globalSchema, EJBSchema ejbSchema, SQLSchema sqlSchema) {
        this.globalSchema = globalSchema;
        this.ejbSchema = ejbSchema;
        this.sqlSchema = sqlSchema;

        queryBuilder = new EJBQueryBuilder(new IdentityDefinerBuilder(ejbSchema, globalSchema));
        mapper = new SchemaMapper(sqlSchema);
    }
    
    public void build() throws QueryException {
        IdentityHashMap associationToRelationshipSlot = new IdentityHashMap();
        IdentityHashMap associationToMTMEndSlots = new IdentityHashMap();
        
        for (Iterator iter = ejbSchema.getEntities().entrySet().iterator(); iter.hasNext();) {
            Map.Entry entry = (Map.Entry) iter.next();
            String name = (String) entry.getKey();
            EJB ejb = (EJB) entry.getValue();
            processEJB(associationToRelationshipSlot, associationToMTMEndSlots, name, ejb);
        }
    }

    private void processEJB(IdentityHashMap associationToRelationshipSlot, IdentityHashMap associationToMTMEndSlots, String name, EJB ejb) throws QueryException {
        if (ejb.isVirtual()) {
            return;
        }
        
        CacheSlot[] slots = processAttributes(ejb);

        Set pkJoinDefinitions = new HashSet();
        Set fkJoinDefinitions = new HashSet();
        EndSlot[] endSlots = processAssociationEnds(associationToRelationshipSlot, associationToMTMEndSlots, ejb, pkJoinDefinitions, fkJoinDefinitions);

        Table table = sqlSchema.getTable(ejb.getName());
        UpdateCommandBuilder commandBuilder = table.getCommandBuilder();
        if (null == commandBuilder) {
            commandBuilder = new StaticCommandBuilder(ejb.getName(), ejbSchema, sqlSchema, globalSchema);
        }
        UpdateCommand createCommand = commandBuilder.buildCreate();
        UpdateCommand storeCommand = commandBuilder.buildStore();
        UpdateCommand removeCommand = commandBuilder.buildRemove();

        CacheTable cacheTable = new CacheTable(name, slots, endSlots, createCommand, storeCommand, removeCommand);
        
        defineJoinDefinitions(pkJoinDefinitions, fkJoinDefinitions, cacheTable);

        globalSchema.addCacheTable(cacheTable);
    }

    private CacheSlot[] processAttributes(EJB ejb) {
        List attributes = ejb.getAttributes();
        
        CacheSlot[] slots = new CacheSlot[attributes.size()];
        for (int i = 0; i < attributes.size(); i++) {
            Attribute attr = (Attribute) attributes.get(i);
            slots[i] = new CacheSlot(attr.getName(), attr.getType(), getDefault(attr.getType()));
        }
        return slots;
    }

    private EndSlot[] processAssociationEnds(IdentityHashMap associationToRelationshipSlot, IdentityHashMap associationToMTMEndSlots, EJB ejb, Set pkJoinDefinitions, Set fkJoinDefinitions) throws QueryException {
        List associationEnds = ejb.getAssociationEnds();
        
        EndSlot[] endSlots = new EndSlot[associationEnds.size()];
        for (int i = 0; i < associationEnds.size(); i++) {
            AssociationEnd end = (AssociationEnd) associationEnds.get(i);
            Association association = end.getAssociation();
            
            RelationshipSlot relationshipSlot = (RelationshipSlot) associationToRelationshipSlot.get(association);
            
            if ( null == relationshipSlot && end.isManyToMany() ) {
                Entity mtmEntity = association.getManyToManyEntity();
                AssociationEnd ends[] = association.getAssociationEnds();
                RelationshipSlot.MutableJoinDefinition leftJoinDefinition = new RelationshipSlot.MutableJoinDefinition();
                RelationshipSlot leftRelationshipSlot = new RelationshipSlot(leftJoinDefinition);
                RelationshipSlot.MutableJoinDefinition rightJoinDefinition = new RelationshipSlot.MutableJoinDefinition();
                RelationshipSlot rightRelationshipSlot = new RelationshipSlot(rightJoinDefinition);
                EndSlot[] mtmSlots = new EndSlot[] {
                        new EndSlot(ends[0].getName(), true, ends[0].isCascadeDelete(), leftRelationshipSlot, true, false),
                        new EndSlot(ends[1].getName(), true, ends[0].isCascadeDelete(), rightRelationshipSlot, true, false)};
                UpdateCommand mtmCreate = mapper.transform(queryBuilder.buildMTMCreate(association));
                UpdateCommand mtmRemove = mapper.transform(queryBuilder.buildMTMRemove(association));

                CacheTable mtmTable = new CacheTable(mtmEntity.getName(), new CacheSlot[0], mtmSlots, mtmCreate,null, mtmRemove);
                leftJoinDefinition.setFKEntity(mtmTable);
                rightJoinDefinition.setFKEntity(mtmTable);

                globalSchema.addCacheTable(mtmTable);
                
                relationshipSlot = new RelationshipSlot(mtmTable, leftJoinDefinition, rightJoinDefinition);
                associationToRelationshipSlot.put(association, relationshipSlot);
                associationToMTMEndSlots.put(association, mtmSlots);
            } else if ( null == relationshipSlot ) {
                relationshipSlot = new RelationshipSlot(new RelationshipSlot.MutableJoinDefinition());
                associationToRelationshipSlot.put(association, relationshipSlot);
            }
            
            if ( end.isManyToMany() ) {
                if ( ejb == association.getLeftJoinDefinition().getPKEntity() ) {
                    pkJoinDefinitions.add(relationshipSlot.getLeftJoinDefinition());
                } else if ( ejb == association.getRightJoinDefinition().getPKEntity() ) {
                    pkJoinDefinitions.add(relationshipSlot.getRightJoinDefinition());
                } else {
                    throw new AssertionError();
                }
            } else {
                if ( ejb == association.getJoinDefinition().getPKEntity() ) {
                    pkJoinDefinitions.add(relationshipSlot.getJoinDefinition());
                } else if ( ejb == association.getJoinDefinition().getFKEntity() ) {
                    fkJoinDefinitions.add(relationshipSlot.getJoinDefinition());
                } else {
                    throw new AssertionError();
                }
            }

            endSlots[i] = new EndSlot(end.getName(), end.isSingle(), end.isCascadeDelete(), relationshipSlot, end.isVirtual(), end.isOnPKSide());
            
            if ( null != relationshipSlot && end.isManyToMany() ) {
                EndSlot[] mtmSlots = (EndSlot[]) associationToMTMEndSlots.get(association);
                if ( mtmSlots[0].getName().equals(end.getName()) ) {
                    ((RelationshipSlot) mtmSlots[0].getAssociation()).addAssociationEnd(endSlots[i]);
                } else if ( mtmSlots[1].getName().equals(end.getName()) ) {
                    ((RelationshipSlot) mtmSlots[1].getAssociation()).addAssociationEnd(endSlots[i]);
                } else {
                    throw new AssertionError();
                }
            }
        }
        return endSlots;
    }

    private void defineJoinDefinitions(Set pkJoinDefinitions, Set fkJoinDefinitions, CacheTable cacheTable) {
        for (Iterator iterator = fkJoinDefinitions.iterator(); iterator.hasNext();) {
            RelationshipSlot.MutableJoinDefinition joinDefinition = (RelationshipSlot.MutableJoinDefinition) iterator.next();
            joinDefinition.setFKEntity(cacheTable);
        }
        
        for (Iterator iterator = pkJoinDefinitions.iterator(); iterator.hasNext();) {
            RelationshipSlot.MutableJoinDefinition joinDefinition = (RelationshipSlot.MutableJoinDefinition) iterator.next();
            joinDefinition.setPKEntity(cacheTable);
        }
    }

    private static final Map DEFAULTS = new HashMap();

    static {
        DEFAULTS.put(Boolean.TYPE, Boolean.FALSE);
        DEFAULTS.put(Byte.TYPE, new Byte((byte) 0));
        DEFAULTS.put(Short.TYPE, new Short((short) 0));
        DEFAULTS.put(Integer.TYPE, new Integer(0));
        DEFAULTS.put(Long.TYPE, new Long(0L));
        DEFAULTS.put(Float.TYPE, new Float(0.0f));
        DEFAULTS.put(Double.TYPE, new Double(0.0d));
        DEFAULTS.put(Character.TYPE, new Character(Character.MIN_VALUE));
    }

    private static Object getDefault(Class type) {
        // assumes get returns null and that is valid ...
        return DEFAULTS.get(type);
    }
    
}
