/**
 *
 *  Copyright 2004 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.builder;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.tranql.cache.CacheTable;
import org.tranql.cache.GlobalSchema;
import org.tranql.ejb.CompoundPKTransform;
import org.tranql.ejb.EJB;
import org.tranql.ejb.EJBSchema;
import org.tranql.ejb.SimplePKTransform;
import org.tranql.field.FieldTransform;
import org.tranql.identity.DerivedIdentity;
import org.tranql.identity.IdentityDefiner;
import org.tranql.identity.IdentityTransform;
import org.tranql.identity.UserDefinedIdentity;
import org.tranql.schema.Attribute;
import org.tranql.schema.Entity;

/**
 *
 *
 *
 * @version $Revision: 323 $ $Date: 2006-02-19 01:29:24 -0500 (Sun, 19 Feb 2006) $
 */
public final class IdentityDefinerBuilder {
    private final EJBSchema ejbSchema;
    private final GlobalSchema globalSchema;
    
    public IdentityDefinerBuilder(EJBSchema ejbSchema, GlobalSchema globalSchema) {
        this.ejbSchema = ejbSchema;
        this.globalSchema = globalSchema;
    }
    
    public EJBSchema getEJBSchema() {
        return ejbSchema;
    }
    
    public IdentityDefiner getIdentityDefiner(Entity entity) {
        List attributes = entity.getAttributes();
        List pkFields = entity.getPrimaryKeyFields();
        int slots[] = new int[pkFields.size()];
        for (int i = 0; i < slots.length; i++) {
            slots[i] = attributes.indexOf(pkFields.get(i));
        }

        CacheTable cacheTable = (CacheTable) globalSchema.getEntity(entity.getName());

        EJB ejb = ejbSchema.getEJB(entity.getName());
        boolean compoundPK = ejb.isCompoundPK();
        if (compoundPK) {
            return new DerivedIdentity(cacheTable, slots);
        } else {
            return new UserDefinedIdentity(cacheTable, slots[0]);
        }
    }

    public IdentityDefiner getIdentityDefiner(Entity entity, int baseIdx) {
        List pkFields = entity.getPrimaryKeyFields();
        int slots[] = new int[pkFields.size()];
        for (int i = 0; i < slots.length; i++) {
            slots[i] = baseIdx + i;
        }

        CacheTable cacheTable = (CacheTable) globalSchema.getEntity(entity.getName());

        EJB ejb = ejbSchema.getEJB(entity.getName());
        boolean compoundPK = ejb.isCompoundPK();
        if (compoundPK) {
            return new DerivedIdentity(cacheTable, slots);
        } else {
            return new UserDefinedIdentity(cacheTable, slots[0]);
        }
    }

    public IdentityDefiner getIdentityDefiner(Entity entity, FieldTransform[] fieldTransforms) {
        List pkFields = entity.getPrimaryKeyFields();
        if (pkFields.size() != fieldTransforms.length) {
            throw new IllegalArgumentException("Wrong number of FieldTransform. Expected [" + pkFields.size() + "]");
        }

        CacheTable cacheTable = (CacheTable) globalSchema.getEntity(entity.getName());

        EJB ejb = (EJB) ejbSchema.getEntity(entity.getName());
        boolean compoundPK = ejb.isCompoundPK();
        if (compoundPK) {
            return new DerivedIdentity(cacheTable, fieldTransforms);
        } else {
            return new UserDefinedIdentity(cacheTable, fieldTransforms[0]);
        }
    }

    public IdentityTransform getPrimaryKeyTransform(Entity entity) {
        List pkFields = entity.getPrimaryKeyFields();
        
        CacheTable cacheTable = (CacheTable) globalSchema.getEntity(entity.getName());

        EJB ejb = ejbSchema.getEJB(entity.getName());
        boolean compoundPK = ejb.isCompoundPK();
        if (false == compoundPK) {
            // single field primary key
            return new SimplePKTransform(cacheTable);
        }

        List fieldNames = new ArrayList();
        pkFields = ejb.getPrimaryKeyFields();
        for (Iterator iter = pkFields.iterator(); iter.hasNext();) {
            Attribute attribute = (Attribute) iter.next();
            fieldNames.add(attribute.getPhysicalName());
        }
        // compound primary key
        Class pkClass = ejb.getPrimaryKeyClass();
        return new CompoundPKTransform(cacheTable, pkClass, fieldNames);
    }

}
