/**
 *
 *  Copyright 2004 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.cache;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;

import org.tranql.abstractschema.AbstractEntity;
import org.tranql.identity.GlobalIdentity;
import org.tranql.intertxcache.CacheFactory;
import org.tranql.intertxcache.NoOpCacheFactory;
import org.tranql.query.UpdateCommand;

/**
 *
 *
 *
 * @version $Revision: 323 $ $Date: 2006-02-19 01:29:24 -0500 (Sun, 19 Feb 2006) $
 */
public final class CacheTable extends AbstractEntity {
    private final List attributes;
    private final List relationships;
    
    private final UpdateCommand insertCommand;
    private final UpdateCommand updateCommand;
    private final UpdateCommand deleteCommand;

    private CacheFactory cacheFactory = NoOpCacheFactory.SINGLETON;
    
    public CacheTable(String name, CacheSlot[] slots, EndSlot[] relationshipSlots, UpdateCommand insertCommand, UpdateCommand updateCommand, UpdateCommand deleteCommand) {
        super(name, name);
        this.insertCommand = insertCommand;
        this.updateCommand = updateCommand;
        this.deleteCommand = deleteCommand;
        
        for (int i = 0; i < slots.length; i++) {
            addAttribute(slots[i]);
        }
        
        if ( null != relationshipSlots ) {
            for (int i = 0; i < relationshipSlots.length; i++) {
                addAssociationEnd(relationshipSlots[i]);
                relationshipSlots[i].setHoldingCacheTable(this);
            }
        }

        attributes = Collections.unmodifiableList(getAttributes());
        relationships = Collections.unmodifiableList(getAssociationEnds());
    }

    public CacheRow newRow() {
        int attributeSize = attributes.size();
        int size = attributeSize + relationships.size();
        Object[] values = new Object[size];
        for (int i = 0; i < attributeSize; i++) {
            CacheSlot slot = (CacheSlot) attributes.get(i);
            values[i] = slot.getDefault();
        }
        
        for (int i = 0; i < relationships.size(); i++) {
            EndSlot relationshipSlot = (EndSlot) relationships.get(i);
            values[attributeSize + i] = relationshipSlot.isMulti()? new HashSet() : null;
        }
        return new CacheRow(values);
    }

    public CacheRow emptyRow(GlobalIdentity id) {
        Object[] noData = new Object[attributes.size() + relationships.size()];
        Arrays.fill(noData, CacheRow.NO_DATA);
        return new CacheRow(id, noData);
    }

    public CacheRow addRow(InTxCache cache, GlobalIdentity id, CacheRow origRow) throws DuplicateIdentityException {
        CacheRow newRow = new CacheRow(id, origRow);
        try {
            cache.associate(newRow);
        } catch (AlreadyAssociatedException e) {
            throw new DuplicateIdentityException(e);
        }
        return newRow;
    }

    public void removeRow(InTxCache cache, GlobalIdentity id) {
        CacheRow row = cache.get(id);
        if (row != null) {
            row.markRemoved();
        }
    }

    public UpdateCommand getInsertCommand() {
        return insertCommand;
    }

    public UpdateCommand getUpdateCommand() {
        return updateCommand;
    }

    public UpdateCommand getDeleteCommand() {
        return deleteCommand;
    }

    public CacheFactory getCacheFactory() {
        return cacheFactory;
    }

    public void setCacheFactory(CacheFactory cacheFactory) {
        this.cacheFactory = cacheFactory;
    }

    public boolean equals(Object obj) {
        // see GlobalIdentity.equals(Object);
        return this == obj;
    }
}
