/**
 *
 *  Copyright 2005 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.ddl;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.tranql.sql.Column;
import org.tranql.sql.EndTable;
import org.tranql.sql.FKColumn;
import org.tranql.sql.SQLSchema;
import org.tranql.sql.Table;


/**
 *
 * @version $Revision: 252 $ $Date: 2005-09-10 10:07:08 -0400 (Sat, 10 Sep 2005) $
 */
public class DDLBuilder {
    
    private final SQLSchema schema;
    
    public DDLBuilder(SQLSchema schema) {
        this.schema = schema;
    }
    
    public DDL createCreate(String entity) {
        Table table = schema.getTable(entity);
        if (null == table) {
            throw new IllegalArgumentException("Entity " + entity + " is undefined.");
        }
        
        Create create = new Create(table);
        
        List columns = table.getAttributes();
        for (Iterator iter = columns.iterator(); iter.hasNext();) {
            Column column = (Column) iter.next();
            if (null != table.getAssociationEndDefiningFKAttribute(column.getPhysicalName())) {
                continue;
            }
            create.addChild(new ColumnDefinition(column, true));
        }
        
        List ends = table.getAssociationEnds();
        for (Iterator iter = ends.iterator(); iter.hasNext();) {
            EndTable endTable = (EndTable) iter.next();
            if (endTable.isOnPKSide()) {
                continue;
            }
            LinkedHashMap pkToFk = endTable.getAssociation().getJoinDefinition().getPKToFKMapping();
            for (Iterator iterator = pkToFk.entrySet().iterator(); iterator.hasNext();) {
                Map.Entry entry = (Map.Entry) iterator.next();
                FKColumn column = (FKColumn) entry.getValue();
                create.addChild(new FKColumnDefinition(column, true));
            }
        }
        
        Constraint constraint = new Constraint(Constraint.PRIMARY_KEY, "PK_" + table.getPhysicalName());
        create.addChild(constraint);
        NodeList nodeList = new NodeList();
        constraint.addChild(nodeList);
        List pkFields = table.getPrimaryKeyFields();
        for (Iterator iter = pkFields.iterator(); iter.hasNext();) {
            Column column = (Column) iter.next();
            nodeList.addChild(new ColumnDefinition(column, false));
        }
        
        ends = table.getAssociationEnds();
        for (Iterator iter = ends.iterator(); iter.hasNext();) {
            EndTable endTable = (EndTable) iter.next();
            if (false == endTable.isOneToOne() || endTable.isOnPKSide()) {
                continue;
            }
            Table relatedTable = (Table) endTable.getEntity();

            constraint = new Constraint(Constraint.UNIQUE_KEY, "UK_" + relatedTable.getPhysicalName());
            create.addChild(constraint);
            nodeList = new NodeList();
            constraint.addChild(nodeList);
            LinkedHashMap pkToFk = endTable.getAssociation().getJoinDefinition().getPKToFKMapping();
            for (Iterator iterator = pkToFk.entrySet().iterator(); iterator.hasNext();) {
                Map.Entry entry = (Map.Entry) iterator.next();
                FKColumn column = (FKColumn) entry.getValue();
                nodeList.addChild(new FKColumnDefinition(column, true));
            }
        }
        
        DDL ddl = new DDL();
        ddl.addChild(create);
        return ddl;
    }

    public DDL[] createConstraints(String entity) {
        Table table = schema.getTable(entity);
        if (null == table) {
            throw new IllegalArgumentException("Entity " + entity + " is undefined.");
        }
        
        List constraints = new ArrayList();
        
        List ends = table.getAssociationEnds();
        for (Iterator iter = ends.iterator(); iter.hasNext();) {
            EndTable endTable = (EndTable) iter.next();
            if (endTable.isOnPKSide()) {
                continue;
            }
            Table relatedTable = (Table) endTable.getEntity();

            Alter alter = new Alter(Alter.ADD, table);
            Constraint constraint = new Constraint(Constraint.FOREIGN_KEY, "FK_" + endTable.getName());
            alter.addChild(constraint);
            NodeList nodeList = new NodeList();
            constraint.addChild(nodeList);
            References references = new References(relatedTable);
            constraint.addChild(references);
            NodeList referencedColumns = new NodeList();
            references.addChild(referencedColumns);
            LinkedHashMap pkToFk = endTable.getAssociation().getJoinDefinition().getPKToFKMapping();
            for (Iterator iterator = pkToFk.entrySet().iterator(); iterator.hasNext();) {
                Map.Entry entry = (Map.Entry) iterator.next();
                Column column = (Column) entry.getKey();
                FKColumn fkColumn = (FKColumn) entry.getValue();
                nodeList.addChild(new FKColumnDefinition(fkColumn, false));
                referencedColumns.addChild(new ColumnDefinition(column, false));
            }
            
            DDL ddl = new DDL();
            ddl.addChild(alter);
            constraints.add(ddl);
        }
        
        return (DDL[]) constraints.toArray(new DDL[constraints.size()]);
    }

    public DDL createDrop(String entity) {
        Table table = schema.getTable(entity);
        if (null == table) {
            throw new IllegalArgumentException("Entity " + entity + " is undefined.");
        }

        DDL ddl = new DDL();
        ddl.addChild(new Drop(table));
        return ddl;
    }
}
