/**
 *
 *  Copyright 2005 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.ddl;

import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.TreeSet;

import org.tranql.schema.EntityComparator;
import org.tranql.sql.SQLSchema;
import org.tranql.sql.Table;

/**
 *
 * @version $Revision: 252 $ $Date: 2005-09-10 10:07:08 -0400 (Sat, 10 Sep 2005) $
 */
public class DDLGenerator {
    private final SQLSchema schema;
    private final DDLCommandBuilder commandBuilder;
    
    public DDLGenerator(SQLSchema schema, DDLCommandBuilder commandBuilder) {
        this.schema = schema;
        this.commandBuilder = commandBuilder;
    }

    public void generate(GenerationStrategy strategy) throws DDLException, IOException {
        Collection tables = schema.getEntities().values();
        
        TreeSet orderedTables = new TreeSet(new EntityComparator());
        orderedTables.addAll(tables);
        
        DDLBuilder builder = new DDLBuilder(schema);
        List ddls = new ArrayList();
        for (Iterator iter = orderedTables.iterator(); iter.hasNext();) {
            Table table = (Table) iter.next();
            strategy.addDDLs(builder, table.getName(), ddls);
        }
        
        List ddlCommands = new ArrayList();
        for (Iterator iter = ddls.iterator(); iter.hasNext();) {
            DDL ddl = (DDL) iter.next();
            ddlCommands.add(commandBuilder.createDDLCommand(ddl));
        }
        
        for (Iterator iter = ddlCommands.iterator(); iter.hasNext();) {
            DDLCommand ddlCommand = (DDLCommand) iter.next();
            strategy.execute(ddlCommand);
        }
    }

    public static interface GenerationStrategy  {
        public void addDDLs(DDLBuilder builder, String entity, List ddls);
        public void execute(DDLCommand command) throws DDLException, IOException;
    }

    public static abstract class AbstractStrategy implements GenerationStrategy {
        private final ExecutionStrategy strategy;

        public AbstractStrategy(ExecutionStrategy strategy) {
            this.strategy = strategy;
        }

        public final void execute(DDLCommand command) throws DDLException, IOException {
            strategy.execute(command);
        }
    }
    
    public static final class CreateStrategy extends AbstractStrategy {
        public CreateStrategy(ExecutionStrategy strategy) {
            super(strategy);
        }

        public void addDDLs(DDLBuilder builder, String entity, List ddls) {
            ddls.add(builder.createCreate(entity));
        }
    }

    public static final class CreateConstraintStrategy extends AbstractStrategy {
        public CreateConstraintStrategy(ExecutionStrategy strategy) {
            super(strategy);
        }

        public void addDDLs(DDLBuilder builder, String entity, List ddls) {
            DDL[] constraints = builder.createConstraints(entity);
            for (int i = 0; i < constraints.length; i++) {
                ddls.add(constraints[i]);
            }
        }
    }

    public static final class DropStrategy extends AbstractStrategy {
        public DropStrategy(ExecutionStrategy strategy) {
            super(strategy);
        }

        public void addDDLs(DDLBuilder builder, String entity, List ddls) {
            ddls.add(builder.createDrop(entity));
        }
    }

    public static final class SequenceStrategy extends AbstractStrategy {
        private final GenerationStrategy strategies[];
        
        public SequenceStrategy(GenerationStrategy strategies[], ExecutionStrategy strategy) {
            super(strategy);
            this.strategies = strategies;
        }

        public void addDDLs(DDLBuilder builder, String entity, List ddls) {
            for (int i = 0; i < strategies.length; i++) {
                strategies[i].addDDLs(builder, entity, ddls);
            }
        }
    }

    public static interface ExecutionStrategy {
        public void execute(DDLCommand command) throws DDLException, IOException;
    }
    
    public static final class WriterExecutionStrategy implements ExecutionStrategy {
        private final Writer writer;

        public WriterExecutionStrategy(Writer writer) {
            this.writer = writer;
        }
        
        public void execute(DDLCommand command) throws DDLException, IOException {
            command.printTo(writer);
            writer.write("\n");
        }
    }
    
    public static final class DBExecutionStrategy implements ExecutionStrategy {
        
        public void execute(DDLCommand command) throws DDLException, IOException {
            command.execute();
        }
    }
}
