/**
 *
 *  Copyright 2005 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.ddl;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.Time;
import java.sql.Timestamp;
import java.sql.Types;
import java.util.Date;

import javax.sql.DataSource;

import org.tranql.sql.Column;
import org.tranql.sql.FKColumn;


/**
 *
 *
 * @version $Revision: 246 $ $Date: 2005-08-26 22:09:42 -0400 (Fri, 26 Aug 2005) $
 */
public class DerbyDDLCommandBuilder implements DDLVisitor, DDLCommandBuilder {
    private final DataTypeMappingRegistry mappingRegistry;
    private final DataSource ds;
    
    public DerbyDDLCommandBuilder(DataSource ds) {
        this.ds = ds;
        
        BaseDataTypeMappingRegistry registry = new BaseDataTypeMappingRegistry();
        registry.register(Types.BIGINT, "BIGINT");
        registry.register(Types.BLOB, "BLOB({2})");
        registry.register(Types.BOOLEAN, "BIT");
        registry.register(Types.CHAR, "CHAR({2})");
        registry.register(Types.BINARY, "CHAR({2}) FOR BIT DATA");
        registry.register(Types.DATE, "DATE");
        registry.register(Types.DECIMAL, "DECIMAL({0}, {1})");
        registry.register(Types.DOUBLE, "DOUBLE");
        registry.register(Types.INTEGER, "INTEGER");
        registry.register(Types.LONGVARCHAR, "LONG VARCHAR");
        registry.register(Types.NUMERIC, "NUMERIC({0}, {1})");
        registry.register(Types.REAL, "REAL");
        registry.register(Types.SMALLINT, "SMALLINT");
        registry.register(Types.TIME, "TIME");
        registry.register(Types.TIMESTAMP, "TIMESTAMP");
        registry.register(Types.VARCHAR, "VARCHAR({2})");
        registry.register(Types.VARBINARY, "VARCHAR({2}) FOR BIT DATA");
        
        registry.register(byte.class, "TYNIINT");
        registry.register(Byte.class, "TYNIINT");
        registry.register(short.class, "SMALLINT");
        registry.register(Short.class, "SMALLINT");
        registry.register(int.class, "INTEGER");
        registry.register(Integer.class, "INTEGER");
        registry.register(BigInteger.class, "INTEGER");
        registry.register(long.class, "BIGINT");
        registry.register(Long.class, "BIGINT");
        registry.register(float.class, "REAL");
        registry.register(Float.class, "REAL");
        registry.register(double.class, "DOUBLE");
        registry.register(Double.class, "DOUBLE");
        registry.register(BigDecimal.class, "NUMERIC({0}, {1})");
        registry.register(boolean.class, "BIT");
        registry.register(Boolean.class, "BIT");
        registry.register(String.class, "VARCHAR({2})");
        registry.register(byte[].class, "VARCHAR({2}) FOR BIT DATA");
        registry.register(Date.class, "DATE");
        registry.register(java.sql.Date.class, "DATE");
        registry.register(Time.class, "TIME");
        registry.register(java.sql.Time.class, "TIME");
        registry.register(Timestamp.class, "TIMESTAMP");
        registry.registerDefault("BLOB({2})");

        mappingRegistry = registry;
    }
    
    public static class Context {
        private final StringBuffer buffer;

        public Context() {
            this.buffer = new StringBuffer();
        }

        public Context(StringBuffer buffer) {
            this.buffer = buffer;
        }

        public Context append(String string) {
            buffer.append(string);
            return this;
        }

        public Context append(char c) {
            buffer.append(c);
            return this;
        }

        public String getText() {
            return buffer.toString();
        }
    }

    public DDLCommand createDDLCommand(DDL ddl) throws DDLException {
        Context ctx = new Context();
        visit(ddl, ctx);
        return new DDLCommand(ddl, ds, ctx.getText());
    }

    public Object visit(DDL node, Object param) throws DDLException {
        for (Node child = node.getChild(); child != null; child = child.getSibling()) {
            param = child.visit(this, param);
        }
        return param;
    }

    public Object visit(Alter node, Object param) throws DDLException {
        Context ctx = (Context) param;
        ctx.append("ALTER TABLE ").append(node.getEntity().getPhysicalName()).append(" ");
        if (Alter.ADD == node.getType()) {
            ctx.append("ADD ");
            node.getChild().visit(this, param);
        }
        return param;
    }
    
    public Object visit(Create node, Object param) throws DDLException {
        Context ctx = (Context) param;
        ctx.append("CREATE TABLE ").append(node.getEntity().getPhysicalName()).append(" (");
        visitChildren(node, param, ", ");
        ctx.append(')');
        return param;
    }

    public Object visit(ColumnDefinition node, Object param) throws DDLException {
        Context ctx = (Context) param;
        Column column = node.getColumn();
        ctx.append(column.getPhysicalName());

        if (node.isDefineSQL()) {
            String type;
            if (column.isColumnDefinitionSet()) {
                type = column.getColumnDefinition();
            } else if (column.isSQLTypeSet()) {
                type = mappingRegistry.getDataType(column.getSQLType(),
                        column.getPrecision(),
                        column.getScale(),
                        column.getLength());
            } else {
                type = mappingRegistry.getDataType(column.getType(),
                        column.getPrecision(),
                        column.getScale(),
                        column.getLength());
            }
            ctx.append(' ').append(type);
        }
        
        return param;
    }
    
    public Object visit(FKColumnDefinition node, Object param) throws DDLException {
        Context ctx = (Context) param;
        FKColumn fkColumn = node.getFKColumn();
        ctx.append(fkColumn.getPhysicalName());
        
        if (node.isDefineSQL()) {
            String type;
            if (fkColumn.isColumnDefinitionSet()) {
                type = fkColumn.getColumnDefinition();
            } else if (fkColumn.isSQLTypeSet()) {
                type = mappingRegistry.getDataType(fkColumn.getSQLType(),
                        fkColumn.getPrecision(),
                        fkColumn.getScale(),
                        fkColumn.getLength());
            } else {
                type = mappingRegistry.getDataType(fkColumn.getType(),
                        fkColumn.getPrecision(),
                        fkColumn.getScale(),
                        fkColumn.getLength());
            }
            ctx.append(' ').append(type);
        }

        return param;
    }
    
    public Object visit(Constraint node, Object param) throws DDLException {
        Context ctx = (Context) param;
        Node columns = node.getChild(); 
        ctx.append("CONSTRAINT ").append(node.getName());
        if (Constraint.PRIMARY_KEY == node.getType()) {
            ctx.append(" PRIMARY KEY");
            columns.visit(this, param);
        } else if (Constraint.FOREIGN_KEY == node.getType()) {
            ctx.append(" FOREIGN KEY");
            columns.visit(this, param);
            Node references = columns.getSibling();
            ctx.append(' ');
            references.visit(this, param);
        } else if (Constraint.UNIQUE_KEY == node.getType()) {
            ctx.append(" UNIQUE");
            columns.visit(this, param);
        }
        return null;
    }
    
    public Object visit(NodeList node, Object param) throws DDLException {
        Context ctx = (Context) param;
        ctx.append('(');
        visitChildren(node, param, ", ");
        ctx.append(')');
        return param;
    }

    public Object visit(References node, Object param) throws DDLException {
        Context ctx = (Context) param;
        ctx.append("REFERENCES ").append(node.getTable().getPhysicalName());
        node.getChild().visit(this, param);
        return param;
    }
    
    public Object visit(Drop node, Object param) throws DDLException {
        Context ctx = (Context) param;
        ctx.append("DROP TABLE ").append(node.getEntity().getPhysicalName());
        return param;
    }
    
    private void visitChildren(Node node, Object param, String sep) throws DDLException {
        Context ctx = (Context) param;
        Node child = node.getChild();
        if (child != null) {
            child.visit(this, param);
            while ((child = child.getSibling()) != null) {
                ctx.append(sep);
                child.visit(this, param);
            }
        }
    }
}
