/**
 *
 *  Copyright 2004 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.ejb;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import org.tranql.abstractschema.AbstractEntity;
import org.tranql.pkgenerator.PrimaryKeyGenerator;
import org.tranql.schema.Attribute;

/**
 *
 *
 * @version $Revision: 431 $ $Date: 2006-11-13 18:18:35 -0500 (Mon, 13 Nov 2006) $
 */
public final class EJB extends AbstractEntity {
    private final EJBProxyFactory proxyFactory;

    private final List virtualCmpFields = new ArrayList();

    private final Class pkClass;
    private final PrimaryKeyGenerator keyGenerator;
    private final boolean virtual;
    private final boolean unknownPK;
    
    private final Collection finders = new ArrayList();
    private final Collection selects = new ArrayList();
    
    private String prefetchGroup;
    
    private boolean selectForUpdate = false;
    
    public EJB(String ejbName, String schemaName, Class pkClass, EJBProxyFactory proxyFactory, PrimaryKeyGenerator keyGenerator, boolean unknownPK) {
        super(ejbName, schemaName);
        this.pkClass = pkClass;
        this.proxyFactory = proxyFactory;
        this.keyGenerator = keyGenerator;
        this.virtual = false;
        this.unknownPK = unknownPK;
    }
    
    public EJB(String ejbName, String schemaName) {
        super(ejbName, schemaName);
        this.pkClass = null;
        this.proxyFactory = null;
        this.keyGenerator = null;
        this.virtual = true;
        this.unknownPK = false;
    }

    public boolean isCompoundPK() {
        List pkFields = getPrimaryKeyFields();
        if (1 < pkFields.size()) {
            return true;
        }
        Attribute attribute = (Attribute) pkFields.get(0);
        return false == pkClass.equals(attribute.getType());
    }
    
    public Class getPrimaryKeyClass() {
        return pkClass;
    }

    public EJBProxyFactory getProxyFactory() {
        return proxyFactory;
    }

    public void addCMPField(CMPField field) {
        addAttribute(field);
    }

    public void addVirtualCMPField(CMPField field) {
        addAttribute(field);
        virtualCmpFields.add(field);
    }
    
    public List getVirtualCMPFields() {
        return Collections.unmodifiableList(virtualCmpFields);
    }
    
    public void addCMRField(CMRField field) {
        addAssociationEnd(field);
    }

    public boolean isAutomaticPrimaryKey() {
        return null != keyGenerator;
    }
    
    public PrimaryKeyGenerator getPrimaryKeyGenerator() {
        return keyGenerator;
    }

    public boolean isUnknownPK() {
        return unknownPK;
    }
    
    /**
     * Indicates if this EJB is "virtual". It is used to identify an EJB
     * abstracting a MTM entity between two other EJB.
     * 
     * @return true if this instance is a MTM EJB, which does not exist per se.
     */
    public boolean isVirtual() {
        return virtual;
    }

    public void addFinder(FinderEJBQLQuery finderEJBQLQuery) {
        finders.add(finderEJBQLQuery);
    }
    
    public Collection getFinders() {
        return Collections.unmodifiableCollection(finders);
    }
    
    public void addSelect(SelectEJBQLQuery selectEJBQLQuery) {
        selects.add(selectEJBQLQuery);
    }

    public Collection getSelects() {
        return Collections.unmodifiableCollection(selects);
    }
    
    public void setPrefetchGroup(String prefetchGroup) {
        this.prefetchGroup = prefetchGroup;
    }
    
    public String getPrefetchGroup() {
        return prefetchGroup;
    }

	/**
	 * @return the selectForUpdate
	 */
	public boolean isSelectForUpdate() {
		return selectForUpdate;
	}

	/**
	 * @param selectForUpdate the selectForUpdate to set
	 */
	public void setSelectForUpdate(boolean selectForUpdate) {
		this.selectForUpdate = selectForUpdate;
	}
}
