/**
 *
 *  Copyright 2004 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.ejb;

import java.util.Iterator;
import java.util.Set;

import org.tranql.cache.AlreadyAssociatedException;
import org.tranql.cache.CacheRow;
import org.tranql.cache.InTxCache;
import org.tranql.identity.GlobalIdentity;
import org.tranql.identity.IdentityDefiner;

/**
 * 
 * 
 * @version $Revision: 160 $ $Date: 2005-01-07 01:26:36 -0500 (Fri, 07 Jan 2005) $
 */
public final class ManyToOneCMR implements CMPFieldTransform {
    private final CMPFieldTransform next;
    private final IdentityDefiner nextIDDefiner;
    private final CMPFieldTransform related;
    private final IdentityDefiner relatedIDDefiner;

    public ManyToOneCMR(CMPFieldTransform next, IdentityDefiner nextIDDefiner, CMPFieldTransform related, IdentityDefiner relatedIDDefiner) {
        this.next = next;
        this.nextIDDefiner = nextIDDefiner;
        this.related = related;
        this.relatedIDDefiner = relatedIDDefiner;
    }

    /**
     * Wraps the underlying set so that changes can be detected and notifications sent
     * @param cache the current transaction cache
     * @param row the row being accessed
     * @return the underlying set wrapped for change detection
     */
    public Object get(InTxCache cache, CacheRow row) {
        return new ManyToOneRelationSet(cache, row.getId(), (Set) next.get(cache, row));
    }

    /**
     * Not supported.
     * @throws UnsupportedOperationException
     */
    public void set(InTxCache cache, CacheRow row, Object value) {
        throw new UnsupportedOperationException();
    }

    private class ManyToOneRelationSet extends AbstractRelationSet {
        private final InTxCache cache;
        private final GlobalIdentity id;

        public ManyToOneRelationSet(InTxCache cache, GlobalIdentity id, Set idSet) {
            super(cache, relatedIDDefiner, idSet);
            this.cache = cache;
            this.id = id;
        }

        protected void associate(GlobalIdentity b2) {
            CacheRow b2Row = cache.get(b2);
            if (null == b2Row) {
                b2Row = b2.getTable().emptyRow(b2);
                try {
                    cache.associate(b2Row);
                } catch (AlreadyAssociatedException e) {
                    throw new AssertionError(e);
                }
                relatedIDDefiner.injectIdentity(b2Row);
            }
            GlobalIdentity a2 = (GlobalIdentity) related.get(cache, b2Row);
            if (a2 != null) {
                CacheRow a2Row = cache.get(a2);
                if ( null == a2Row ) {
                    a2Row = a2.getTable().emptyRow(a2);
                    try {
                        cache.associate(a2Row);
                    } catch (AlreadyAssociatedException e) {
                        // This can not happen.
                        throw new AssertionError(e);
                    }
                    nextIDDefiner.injectIdentity(a2Row);
                }
                Set a2Set = (Set) next.get(cache, a2Row);
                a2Set.remove(b2);
            }
            related.set(cache, b2Row, id);
        }

        protected void disassociate(GlobalIdentity b2) {
            CacheRow b2Row = cache.get(b2);
            related.set(cache, b2Row, null);
        }

        public Iterator iterator() {
            return idSet.iterator();
        }
    }
}
