/**
 *
 *  Copyright 2005 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.ejbqlcompiler;

import java.util.ArrayList;
import java.util.Collection;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.tranql.builder.QueryTileBuilder;
import org.tranql.builder.QueryTileBuilder.QuerySourceFactory;
import org.tranql.ql.DerivedTable;
import org.tranql.ql.Join;
import org.tranql.ql.Node;
import org.tranql.ql.Path;
import org.tranql.ql.Query;
import org.tranql.ql.QueryException;
import org.tranql.ql.QuerySource;
import org.tranql.ql.QuerySourceDictionary;
import org.tranql.ql.Where;
import org.tranql.schema.AssociationEnd;

/**
 * Transform DerivedTable to cross-join clauses.
 * 
 * @version $Revision: 323 $ $Date: 2006-02-19 01:29:24 -0500 (Sun, 19 Feb 2006) $
 */
class DerivedTableToCrossJoinTransformer extends AbstractSubQueryAwareTransformer {
    private static final DerivedTableSearcher SEARCHER = new DerivedTableSearcher();
    
    public DerivedTableToCrossJoinTransformer(QuerySourceDictionary sourceDictionary) {
        super(sourceDictionary);
    }
    
    public Object visit(Query node, Object param) throws QueryException {
    	QuerySourceDictionary sourceDictionary = (QuerySourceDictionary) param;
        StackNodeRegister register = new StackNodeRegister();  
        SEARCHER.visit(node, register);
        Context ctx = new Context(register.getRegisters(), sourceDictionary);
        
        return super.visit(node, ctx);
    }
    
    public Object visit(DerivedTable node, Object param) throws QueryException {
        Context ctx = (Context) param;
        
        Path path = node.getPath();
        List thePath = path.getPath();
        Node join = null;
        for (int i = 0; i < thePath.size(); i++) {
            AssociationEnd end = (AssociationEnd) thePath.get(i);
            QuerySource entitySource = ctx.factory(i + 1, node);
            if (end.isManyToMany()) {
                QuerySource mtmQS = ctx.factoryMTM(node);
                Node mtmJoin = new Join(Join.CROSS_JOIN).addChild(mtmQS).addChild(entitySource);
                join = null == join ? mtmJoin : new Join(Join.CROSS_JOIN).addChild(join).addChild(mtmJoin);
            } else {
                join = null == join ? entitySource : new Join(Join.CROSS_JOIN).addChild(join).addChild(entitySource);
            }
        }
        
        return join;
    }
    
    public Object visit(Where node, Object param) throws QueryException {
        Context ctx = (Context) param;

        Collection derivedTables = ctx.getRegisteredDerivedTables();
        Collection factories = new ArrayList();
        for (Iterator iter = derivedTables.iterator(); iter.hasNext();) {
            DerivedTable table = (DerivedTable) iter.next();
            factories.add(new DerivedTableFactory(ctx, table));
        }

        Node operation = QueryTileBuilder.buildJoinCondition(factories);

        if (null == operation) {
            return super.visit(node, param);
        }

        return QueryTileBuilder.buildWhere(node, operation, this, param);
    }

    private static class DerivedTableSearcher extends AbstractNodeSearcher {
        private static final Object DERIVED_TABLE = new Object();
        
        public Object visit(DerivedTable node, Object param) throws QueryException {
            StackNodeRegister register = (StackNodeRegister) param;
            register.put(DERIVED_TABLE, node);
            return param;
        }
    }
    
    private static class DerivedTableFactory implements QuerySourceFactory {
        private final Context ctx;
        private final DerivedTable derivedTable;
        
        public DerivedTableFactory(Context ctx, DerivedTable derivedTable) {
            this.ctx = ctx;
            this.derivedTable = derivedTable;
        }
        
        public QuerySource factory(int index) throws QueryException {
            return ctx.factory(index, derivedTable);
        }
        
        public QuerySource factoryMTM() throws QueryException {
            return ctx.factoryMTM(derivedTable);
        }
        
        public Path getPath() {
            return derivedTable.getPath();
        }
    }
    
    private static class Context extends AbstractSubQueryContext {
        private final IdentityHashMap nodeToDerivedTables = new IdentityHashMap();
        private final IdentityHashMap derivedTablesQS = new IdentityHashMap();
        
        private Context(IdentityHashMap registers, QuerySourceDictionary sourceDictionary) throws QueryException {
            for (Iterator iter = registers.entrySet().iterator(); iter.hasNext();) {
                Map.Entry entry = (Map.Entry) iter.next();
                Node node = (Node) entry.getKey();
                StackNodeRegister.NodeRegister register = (StackNodeRegister.NodeRegister) entry.getValue();
                
                List derivedTables = register.get(DerivedTableSearcher.DERIVED_TABLE);
                nodeToDerivedTables.put(node, derivedTables);
                for (int i = 0; i < derivedTables.size(); i++) {
                    DerivedTable derivedTable = (DerivedTable) derivedTables.get(i);

                    Path path = derivedTable.getPath();
                    List thePath = path.getPath();
                    QuerySource[] sources = new QuerySource[thePath.size() + 2];
                    derivedTablesQS.put(derivedTable, sources);

                    QuerySource qs = sourceDictionary.retrieveQuerySource(path.getRoot().getAlias());
                    if (null == qs) {
                        throw new QueryException("Unknown alias=" + path.getRoot().getAlias());
                    }
                    sources[0] = qs;
                    
                    for (int j = 0; j < thePath.size(); j++) {
                        AssociationEnd end = (AssociationEnd) thePath.get(j);
                        if (thePath.size() - 1 == j) {
                            sources[j + 1] = new QuerySource(end.getEntity(), derivedTable.getAlias());
                            sourceDictionary.registerQuerySource(sources[j + 1]);
                            if (end.isManyToMany()) {
                                sources[j + 2] = new QuerySource(end.getAssociation().getManyToManyEntity(), sourceDictionary.getNextAlias());
                            }
                        } else {
                            sources[j + 1] = new QuerySource(end.getEntity(), sourceDictionary.getNextAlias());
                        }
                    }
                }
            }
        }
        
        public Collection getRegisteredDerivedTables() {
            return (Collection) nodeToDerivedTables.get(peek());
        }
        
        public QuerySource factory(int index, DerivedTable derivedTable) throws QueryException {
            QuerySource[] sources = ((QuerySource[]) derivedTablesQS.get(derivedTable));
            if (null == sources) {
                throw new AssertionError("Unknown DerivedTable=" + derivedTable);
            }
            
            return sources[index];
        }

        public QuerySource factoryMTM(DerivedTable derivedTable) throws QueryException {
            QuerySource[] sources = ((QuerySource[]) derivedTablesQS.get(derivedTable));
            if (null == sources) {
                throw new AssertionError("Unknown DerivedTable=" + derivedTable);
            }
            
            return sources[sources.length - 1];
        }

    }
    
}
