/**
 *
 *  Copyright 2005 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.ejbqlcompiler;

import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.tranql.builder.IdentityDefinerBuilder;
import org.tranql.builder.QueryTileBuilder;
import org.tranql.ql.AbstractNode;
import org.tranql.ql.AliasedEntity;
import org.tranql.ql.AttributeReference;
import org.tranql.ql.BinaryOperation;
import org.tranql.ql.EntityReference;
import org.tranql.ql.Node;
import org.tranql.ql.Not;
import org.tranql.ql.ParameterReference;
import org.tranql.ql.Query;
import org.tranql.ql.QueryBinding;
import org.tranql.ql.QueryException;
import org.tranql.ql.QueryVisitor;
import org.tranql.ql.RowConstructor;
import org.tranql.ql.Select;
import org.tranql.schema.Attribute;
import org.tranql.schema.Entity;

/**
 * Query transform that replaces entity references with operations that use
 * just the primary key columns. Primarily intended to support SQL92 databases.
 *
 * @version $Revision: 323 $ $Date: 2006-02-19 01:29:24 -0500 (Sun, 19 Feb 2006) $
 */
public class EntityReferenceTransformer extends InterRepresentationTransformer {
    private final IdentityDefinerBuilder identityDefinerBuilder;
    
    public EntityReferenceTransformer(IdentityDefinerBuilder identityDefinerBuilder) {
        super(null);
        this.identityDefinerBuilder = identityDefinerBuilder;
    }

    public Object visit(Query node, Object param) throws QueryException {
        Context ctx = new Context();
        ctx.query = node;
        ctx.paramTransforms = node.getParamTransforms();

        Node to = new PlaceHolderNode();
        visitChildren(to, node, ctx);
        
        Query query = new Query(ctx.paramTransforms, node.getResultTransforms());
        query.addChild(to.getChild());
        return query;
    }
    
    public Object visit(Select node, Object param) throws QueryException {
        Node to = new Select(node.isDistinct());
        for (Node child = node.getChild(); child != null; child = child.getSibling()) {
            if (child instanceof EntityReference) {
                AliasedEntity alias = ((EntityReference) child).getEntity();
                List pkColumns = alias.getEntity().getPrimaryKeyFields();
                for (int i = 0; i < pkColumns.size(); i++) {
                    Attribute attribute = (Attribute) pkColumns.get(i);
                    to.addChild(new AttributeReference(alias, attribute));
                }
            } else {
                child = (Node) child.visit(this, param);
                to.addChild(child);
            }
        }
        return to;
    }

    public Object visit(EntityReference node, Object param) throws QueryException {
        Node to;
        AliasedEntity alias = node.getEntity();
        Entity entity = alias.getEntity();
        List pkAttributes = entity.getPrimaryKeyFields();
        if (pkAttributes.size() == 1) {
            to = new AttributeReference(alias, (Attribute) pkAttributes.get(0));
        } else {
            to = new RowConstructor();
            for (int i = 0; i < pkAttributes.size(); i++) {
                Attribute column = (Attribute) pkAttributes.get(i);
                to.addChild(new AttributeReference(alias, column));
            }
        }
        return to;
    }

    public Object visit(BinaryOperation node, Object param) throws QueryException {
        String operation = node.getOperation();

        Node left = node.getChild();
        Node right = left.getSibling();
        Node to;
        if (left instanceof EntityReference) {
            if (right instanceof EntityReference) {
                to = compareTwoEntities((EntityReference) left, (EntityReference) right);
            } else if (right instanceof ParameterReference) {
                to = compareEntityToParameter((EntityReference) left, (ParameterReference) right, (Context) param);
            } else {
                throw new UnsupportedOperationException();
            }

            if (BinaryOperation.NE.equals(operation)) {
                to = new Not().addChild(to);
            } else if (!BinaryOperation.EQ.equals(operation)) {
                throw new UnsupportedOperationException();
            }
        } else if (right instanceof EntityReference) {
            if (left instanceof ParameterReference) {
                to = compareEntityToParameter((EntityReference)right, (ParameterReference) left, (Context) param);
            } else {
                throw new UnsupportedOperationException();
            }

            if (BinaryOperation.NE.equals(operation)) {
                to = new Not().addChild(to);
            } else if (!BinaryOperation.EQ.equals(operation)) {
                throw new UnsupportedOperationException();
            }
        } else {
            to = (Node) super.visit(node, param);
        }

        return to;
    }

    private Node compareTwoEntities(EntityReference leftReference, EntityReference rightReference) throws QueryException {
        
        Node to = null;
        AliasedEntity leftAlias = leftReference.getEntity();
        AliasedEntity rightAlias = rightReference.getEntity();
        Entity entity = leftAlias.getEntity();
        if (!entity.equals(rightAlias.getEntity())) {
            throw new QueryException("Cannot compare different entities: " + entity.getName() + " to " + rightAlias.getEntity().getName());
        }
        List keyAttrs = entity.getPrimaryKeyFields();
        for (int i = 0; i < keyAttrs.size(); i++) {
            Attribute leftAttribute = (Attribute) keyAttrs.get(i);
            Node compare = new BinaryOperation(BinaryOperation.EQ);
            compare.addChild(new AttributeReference(leftAlias, leftAttribute));
            compare.addChild(new AttributeReference(rightAlias, leftAttribute));
            to = (to == null) ? compare : new BinaryOperation(BinaryOperation.AND).addChild(to).addChild(compare);
        }
        return to;
    }

    private Node compareEntityToParameter(EntityReference entityReference, ParameterReference paramReference, Context ctx) throws QueryException {
        Node to = null;
        AliasedEntity alias = entityReference.getEntity();
        Entity entity = alias.getEntity();
        
        LinkedHashMap newBindings = QueryTileBuilder.buildQueryBindings(ctx.paramTransforms, paramReference.getParam(), entity, identityDefinerBuilder);
        
        QueryBinding[] newParams = new QueryBinding[ctx.paramTransforms.length + newBindings.size()];
        System.arraycopy(ctx.paramTransforms, 0, newParams, 0, ctx.paramTransforms.length);
        int i = 0;
        for (Iterator iter = newBindings.entrySet().iterator(); iter.hasNext();) {
            Map.Entry entry = (Map.Entry) iter.next();
            ParameterReference parameterReference = (ParameterReference) entry.getKey();
            QueryBinding queryBinding = (QueryBinding) entry.getValue();
            
            Node compare = new BinaryOperation(BinaryOperation.EQ);
            compare.addChild(new AttributeReference(alias, queryBinding.getAttribute()));
            compare.addChild(parameterReference);
            to = (to == null) ? compare : new BinaryOperation(BinaryOperation.AND).addChild(to).addChild(compare);
            
            newParams[ctx.paramTransforms.length + i] = queryBinding;
            i++;
        }
        ctx.paramTransforms = newParams;
        
        return to;
    }
    
    private static class Context {
        private Query query;
        private QueryBinding[] paramTransforms;
    }
    
    private static class PlaceHolderNode extends AbstractNode {
        public Object visit(QueryVisitor visitor, Object param) throws QueryException {
            throw new AssertionError();
        }
    }
}
