/**
 *
 *  Copyright 2005 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.ejbqlcompiler;

import java.util.Collection;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.tranql.builder.IdentityDefinerBuilder;
import org.tranql.builder.QueryTileBuilder;
import org.tranql.ql.AttributeReference;
import org.tranql.ql.BinaryOperation;
import org.tranql.ql.EntityReference;
import org.tranql.ql.From;
import org.tranql.ql.In;
import org.tranql.ql.Join;
import org.tranql.ql.MemberOf;
import org.tranql.ql.Node;
import org.tranql.ql.ParameterReference;
import org.tranql.ql.Path;
import org.tranql.ql.Query;
import org.tranql.ql.QueryBinding;
import org.tranql.ql.QueryException;
import org.tranql.ql.QuerySource;
import org.tranql.ql.QuerySourceDictionary;
import org.tranql.ql.RowConstructor;
import org.tranql.ql.Select;
import org.tranql.ql.SubQuery;
import org.tranql.ql.Where;
import org.tranql.schema.Association;
import org.tranql.schema.AssociationEnd;
import org.tranql.schema.Attribute;

/**
 * Transform MemberOf clauses.
 * 
 * @version $Revision: 323 $ $Date: 2006-02-19 01:29:24 -0500 (Sun, 19 Feb 2006) $
 */
class MemberOfTransformer extends AbstractSubQueryAwareTransformer {
    private static final MemberOfSearcher SEARCHER = new MemberOfSearcher();

    private final IdentityDefinerBuilder identityDefinerBuilder;
    
    public MemberOfTransformer(IdentityDefinerBuilder identityDefinerBuilder, QuerySourceDictionary sourceDictionary) {
        super(sourceDictionary);
        this.identityDefinerBuilder = identityDefinerBuilder;
    }
    
    public Object visit(Query node, Object param) throws QueryException {
    	QuerySourceDictionary sourceDictionary = (QuerySourceDictionary) param;
        StackNodeRegister register = new StackNodeRegister(); 
        SEARCHER.visit(node, register);
        Context ctx = new Context(register.getRegisters(), sourceDictionary);
        ctx.paramTransforms = node.getParamTransforms();

        Node returned = (Node) super.visit(node, ctx);
        
        Query query = new Query(ctx.paramTransforms, node.getResultTransforms());
        query.addChild(returned.getChild());
        return query;
    }
    
    public Object visit(From node, Object param) throws QueryException {
        Context ctx = (Context) param;
        
        Collection memberOfs = ctx.getRegisteredMemberOfs();
        Node to = (From) super.visit(node, param);
        for (Iterator iter = memberOfs.iterator(); iter.hasNext();) {
            MemberOf memberOf = (MemberOf) iter.next();
            Node designatedValue = memberOf.getDesignatedValue();
            if (false == designatedValue instanceof Path) {
                continue;
            }
            
            Path path = (Path) designatedValue;
            List thePath = path.getPath();
            QuerySource previousQS = ctx.factoryLeftPath(0, memberOf);
            for (int i = 0; i < thePath.size(); i++) {
                QuerySource qs = ctx.factoryLeftPath(i + 1, memberOf);
                if (memberOf.isNot()) {
                    AssociationEnd end = (AssociationEnd) thePath.get(i);
                    Association association = end.getAssociation();
                    
                    to =  new Join(Join.LEFT_JOIN).addChild(to).addChild(qs);
                    
                    Association.JoinDefinition joinDefinition = association.getJoinDefinition();
                    Node condition = QueryTileBuilder.buildJoinCondition(joinDefinition, previousQS, qs);
                    to.addChild(condition);
                } else {
                    to =  new Join(Join.CROSS_JOIN).addChild(to).addChild(qs);
                }
            }
        }
    
        return to;
    }
    
    public Object visit(MemberOf node, Object param) throws QueryException {
        Context ctx = (Context) param;

        // a MEMBER OF is replaced by an (...) IN (sub-query).
        In in = new In(node.isNot());
        
        SubQuery subQuery = new SubQuery();

        // builds the from clause of the sub-query.
        Path path = node.getCMRPath();
        List thePath = path.getPath();
        Node fromNode = null;
        QuerySource lastQS = null;
        for (int i = 0; i < thePath.size(); i++) {
            AssociationEnd end = (AssociationEnd) thePath.get(i);
            Node append = null;
            lastQS = ctx.factory(i + 1, node);
            if (end.isManyToMany()) {
                QuerySource mtmQS = ctx.factoryMTM(node);
                append = new Join(Join.CROSS_JOIN).addChild(mtmQS).addChild(lastQS);
            } else {
                append = lastQS;
            }
            fromNode = null == fromNode ? append : new Join(Join.CROSS_JOIN).addChild(fromNode).addChild(append);
        }
        
        // builds the criteria corresponding to the join of the multi-valued cmr-path. 
        Node whereNode = QueryTileBuilder.buildJoinCondition(new MemberOfFactory(ctx, node));

        // selects the primary key fields of the multi-valued cmr-path.
        Select select = new Select(false);
        AssociationEnd end = (AssociationEnd) thePath.get(thePath.size() - 1);
        List pkFields = end.getEntity().getPrimaryKeyFields();
        for (Iterator iterator = pkFields.iterator(); iterator.hasNext();) {
            Attribute attribute = (Attribute) iterator.next();
            select.addChild(new AttributeReference(lastQS, attribute));
        }
        
        Node returnedNode;
        Node subQueryLeft;
        Node designatedValue = node.getDesignatedValue();
        if (designatedValue instanceof Path) {
            Path designatedPath = (Path) designatedValue;
            List designatedThePath = designatedPath.getPath(); 
            AssociationEnd designatedEnd = (AssociationEnd) designatedThePath.get(designatedThePath.size() - 1);
            QuerySource qs = ctx.factoryLeftPath(designatedThePath.size(), node);
            subQueryLeft = new RowConstructor();
            List pkKeys = designatedEnd.getEntity().getPrimaryKeyFields();
            for (int i = 0; i < pkKeys.size(); i++) {
                Attribute column = (Attribute) pkKeys.get(i);
                subQueryLeft.addChild(new AttributeReference(qs, column));
            }

            Node joinDesignatedPath = null;
            
            if (false == node.isNot()) {
                // appends the criteria corresponding to the single valued cmr-path being tested.
                joinDesignatedPath =QueryTileBuilder.buildJoinCondition(new MemberOfCMRPathFactory(ctx, node));
            }

            if (null == joinDesignatedPath) {
                returnedNode = in;
            } else {
                returnedNode = new BinaryOperation(BinaryOperation.AND).addChild(joinDesignatedPath).addChild(in);
                returnedNode = new RowConstructor().addChild(returnedNode);
            }
        } else if (designatedValue instanceof EntityReference) {
            EntityReference entityReference = (EntityReference) designatedValue;
            if (end.getEntity() != entityReference.getEntity().getEntity()) {
                throw new QueryException("Single valued cmr-path is of the wrong type; MemberOf=" + node);
            }
            
            subQueryLeft = new RowConstructor();
            List pkKeys = entityReference.getEntity().getEntity().getPrimaryKeyFields();
            for (int i = 0; i < pkKeys.size(); i++) {
                Attribute column = (Attribute) pkKeys.get(i);
                subQueryLeft.addChild(new AttributeReference(entityReference.getEntity(), column));
            }
            returnedNode = in;
        } else if (designatedValue instanceof ParameterReference) {
            ParameterReference paramReference = (ParameterReference) designatedValue;

            LinkedHashMap newBindings = QueryTileBuilder.buildQueryBindings(ctx.paramTransforms, paramReference.getParam(), end.getEntity(), identityDefinerBuilder);

            subQueryLeft = new RowConstructor();

            QueryBinding[] newParams = new QueryBinding[ctx.paramTransforms.length + newBindings.size()];
            System.arraycopy(ctx.paramTransforms, 0, newParams, 0, ctx.paramTransforms.length);
            int i = 0;
            for (Iterator iter = newBindings.entrySet().iterator(); iter.hasNext();) {
                Map.Entry entry = (Map.Entry) iter.next();
                paramReference = (ParameterReference) entry.getKey();
                QueryBinding queryBinding = (QueryBinding) entry.getValue();

                subQueryLeft.addChild(paramReference);
                
                newParams[ctx.paramTransforms.length + i] = queryBinding;
                i++;
            }
            ctx.paramTransforms = newParams;

            returnedNode = in;
        } else {
            throw new QueryException("Designated value is of the wrong type. MemberOf=" + node);
        }

        // concatenates the IN (sub-query).
        in.addChild(subQueryLeft);
        in.addChild(subQuery);
        subQuery.addChild(select);
        subQuery.addChild(new From().addChild(fromNode));
        subQuery.addChild(new Where().addChild(whereNode));

        return returnedNode;
    }

    private static class MemberOfSearcher extends AbstractNodeSearcher {
        private static final Object MEMBER_OF = new Object();

        public Object visit(MemberOf node, Object param) throws QueryException {
            StackNodeRegister register = (StackNodeRegister) param;
            register.put(MEMBER_OF, node);
            return param;
        }
    }
    
    private static class MemberOfFactory implements QueryTileBuilder.QuerySourceFactory {
        private final Context ctx;
        private final MemberOf memberOf;
        
        public MemberOfFactory(Context ctx, MemberOf memberOf) {
            this.ctx = ctx;
            this.memberOf = memberOf;
        }
        
        public QuerySource factory(int index) throws QueryException {
            return ctx.factory(index, memberOf);
        }

        public QuerySource factoryMTM() throws QueryException {
            return ctx.factoryMTM(memberOf);
        }

        public Path getPath() {
            return memberOf.getCMRPath();
        }
    }

    private static class MemberOfCMRPathFactory implements QueryTileBuilder.QuerySourceFactory {
        private final Context ctx;
        private final MemberOf memberOf;
        
        public MemberOfCMRPathFactory(Context ctx, MemberOf memberOf) {
            this.ctx = ctx;
            this.memberOf = memberOf;
        }
        
        public QuerySource factory(int index) throws QueryException {
            return ctx.factoryLeftPath(index, memberOf);
        }

        public QuerySource factoryMTM() throws QueryException {
            throw new QueryException("Designed valued of a CMR path is a multi-valued cmr-path. MemberOf=" + memberOf);
        }

        public Path getPath() {
            return (Path) memberOf.getDesignatedValue();
        }
    }

    private static class Context extends AbstractSubQueryContext {
        private final QuerySourceDictionary sourceDictionary;
        private final IdentityHashMap nodeToMemberOfs = new IdentityHashMap();
        private final IdentityHashMap memberOfsQS = new IdentityHashMap();
        private final IdentityHashMap memberOfLeftsQS = new IdentityHashMap();
        private QueryBinding[] paramTransforms;
        
        private Context(IdentityHashMap registers, QuerySourceDictionary sourceDictionary) throws QueryException {
            this.sourceDictionary = sourceDictionary;

            for (Iterator iter = registers.entrySet().iterator(); iter.hasNext();) {
                Map.Entry entry = (Map.Entry) iter.next();
                Node node = (Node) entry.getKey();
                StackNodeRegister.NodeRegister register = (StackNodeRegister.NodeRegister) entry.getValue();
            
                List memberOfs = register.get(MemberOfSearcher.MEMBER_OF);
                nodeToMemberOfs.put(node, memberOfs);
                
                for (int i = 0; i < memberOfs.size(); i++) {
                    MemberOf memberOf = (MemberOf) memberOfs.get(i);
                    
                    Path path = memberOf.getCMRPath();
                    List thePath = path.getPath();
                    QuerySource[] sources = new QuerySource[thePath.size() + 2];
                    memberOfsQS.put(memberOf, sources);
                        
                    QuerySource qs = sourceDictionary.retrieveQuerySource(path.getRoot().getAlias());
                    if (null == qs) {
                        throw new QueryException("Unknown alias=" + path.getRoot().getAlias());
                    }
                    sources[0] = qs;
                    
                    for (int j = 0; j < thePath.size(); j++) {
                        AssociationEnd end = (AssociationEnd) thePath.get(j);
                        sources[j + 1] = new QuerySource(end.getEntity(), sourceDictionary.getNextAlias());
                        sourceDictionary.registerQuerySource(sources[j + 1]);
                        if (end.isManyToMany()) {
                            sources[j + 2] = new QuerySource(end.getAssociation().getManyToManyEntity(), sourceDictionary.getNextAlias());
                            sourceDictionary.registerQuerySource(sources[j + 2]);
                        }
                    }
                }
            }
        }
        
        public Collection getRegisteredMemberOfs() {
            return (Collection) nodeToMemberOfs.get(peek());
        }

        public QuerySource factory(int index, MemberOf memberOf) throws QueryException {
            QuerySource[] sources = (QuerySource[]) memberOfsQS.get(memberOf);
            if (null == sources) {
                throw new AssertionError("Unknown MemberOf=" + memberOf);
            }
            
            return sources[index];
        }

        public QuerySource factoryMTM(MemberOf memberOf) throws QueryException {
            QuerySource[] sources = (QuerySource[]) memberOfsQS.get(memberOf);
            if (null == sources) {
                throw new AssertionError("Unknown MemberOf=" + memberOf);
            }
            
            return sources[sources.length - 1];
        }

        public QuerySource factoryLeftPath(int index, MemberOf memberOf) throws QueryException {
            QuerySource[] sources = (QuerySource[]) memberOfLeftsQS.get(memberOf);
            if (null == sources) {
                Path path = (Path) memberOf.getDesignatedValue();
                
                List thePath = path.getPath();
                sources = new QuerySource[thePath.size() + 1];
                memberOfLeftsQS.put(memberOf, sources);

                QuerySource qs = sourceDictionary.retrieveQuerySource(path.getRoot().getAlias());
                if (null == qs) {
                    throw new QueryException("Unknown alias=" + path.getRoot().getAlias());
                }
                sources[0] = qs;
                
                for (int i = 0; i < thePath.size(); i++) {
                    AssociationEnd end = (AssociationEnd) thePath.get(i);
                    sources[i + 1] = new QuerySource(end.getEntity(), sourceDictionary.getNextAlias());
                    sourceDictionary.registerQuerySource(sources[i + 1]);
                }
            }
            
            return sources[index];
        }
        
    }
    
}
