/**
 *
 *  Copyright 2004 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.field;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.ObjectStreamClass;

/**
 * Transform that stores a Serializable object in the cache as a byte array.
 *
 * @version $Revision: 45 $ $Date: 2004-04-19 18:23:20 -0400 (Mon, 19 Apr 2004) $
 */
public class FieldMarshaller implements FieldTransform {
    private final FieldTransform next;
    private final ClassLoader cl;
    private final Class fieldClass;

    public FieldMarshaller(FieldTransform next, ClassLoader cl, Class fieldClass) {
        this.next = next;
        this.cl = cl;
        this.fieldClass = fieldClass;
    }

    public Class getFieldClass() {
        return fieldClass;
    }

    public Object get(Row row) throws FieldTransformException {
        Object value = next.get(row);
        if (value != null) {
            ByteArrayInputStream bais = new ByteArrayInputStream((byte[]) value);
            try {
                ObjectInputStream ois = new MarshalledInputStream(bais);
                try {
                    value = ois.readObject();
                } finally {
                    ois.close();
                }
            } catch (IOException e) {
                throw new FieldTransformException("Unable to unmarshal value", e);
            } catch (ClassNotFoundException e) {
                throw new FieldTransformException("Unable to unmarshal value", e);
            }
        }
        return value;
    }

    public void set(Row row, Object value) throws FieldTransformException {
        if (value != null) {
            try {
                ByteArrayOutputStream baos = new ByteArrayOutputStream();
                ObjectOutputStream oos = new ObjectOutputStream(baos);
                oos.writeObject(value);
                oos.close();
                value = baos.toByteArray();
            } catch (IOException e) {
                throw new FieldTransformException("Unable to marshal value", e);
            }
        }
        next.set(row, value);
    }

    /**
     * Specialization of ObjectInputStream to use a designated classloader.
     */
    private class MarshalledInputStream extends ObjectInputStream {
        public MarshalledInputStream(InputStream in) throws IOException {
            super(in);
        }

        protected Class resolveClass(ObjectStreamClass desc) throws IOException, ClassNotFoundException {
            try {
                return cl.loadClass(desc.getName());
            } catch (ClassNotFoundException e) {
                // let the parent try
                return super.resolveClass(desc);
            }
        }
    }
}
