/**
 * 
 *  Copyright 2005 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.intertxcache;

import java.util.Collection;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.Map;

import javax.transaction.Synchronization;
import javax.transaction.Transaction;

import org.tranql.cache.CacheRow;
import org.tranql.cache.CacheRowState;
import org.tranql.identity.GlobalIdentity;

/**
 *
 * @version $Revision: 341 $ $Date: 2006-06-03 18:16:22 -0400 (Sat, 03 Jun 2006) $
 */
public class TxCache implements Cache {
    private final Cache next;
    private final Map txToCache;
    
    public TxCache(Cache next) {
        this.next = next;
        txToCache = new IdentityHashMap();
    }

    public void transactionStart(Transaction tx) throws CacheException {
        next.transactionStart(tx);
    }
    
    public void transactionEnd(Transaction tx, int status, Collection rows) throws CacheException {
        next.transactionEnd(tx, status, rows);
    }
    
    public void put(Transaction tx, CacheRow row) throws CacheException {
        if (null == tx) {
            next.put(tx, row);
            return;
        }

        Map cache = getTxCache(tx);
        cache.put(row.getId(), new CacheRowEntry(row));
        if (CacheRowState.CLEAN == row.getState()) {
            next.put(tx, row);
        }
    }
    
    public void set(Transaction tx, CacheRow target, int[] indices) throws CacheException {
        if (null == tx) {
            next.set(tx, target, indices);
            return;
        }
        getTxCache(tx);

        int[] tmpIndices = new int[indices.length];
        int tmpIndex = 0;
        for (int i = 0; i < tmpIndices.length; i++) {
            if (target.isModified(i)) {
                continue;
            }
            tmpIndices[tmpIndex++] = indices[i];
        }
        
        if (0 == tmpIndex) {
            return;
        }
        indices = new int[tmpIndex];
        System.arraycopy(tmpIndices, 0, indices, 0, tmpIndex);

        next.set(tx, target, indices);
    }

    public CacheRow get(Transaction tx, GlobalIdentity id) throws CacheException {
        if (null == tx) {
            return next.get(tx, id);
        }
        Map cache = getTxCache(tx);

        CacheRowEntry row = (CacheRowEntry) cache.get(id);
        if (null != row) {
            return row.buildCacheRow();
        }
        
        return next.get(tx, id);
    }
    
    public boolean get(Transaction tx, CacheRow target, int[] indices) throws CacheException {
        if (null == tx) {
            return next.get(tx, target, indices);
        }
        Map cache = getTxCache(tx);
        
        CacheRowEntry row = (CacheRowEntry) cache.get(target.getId());
        if (null != row && row.isRemoved()) {
            throw new IllegalStateException("Entity has been removed.");
        }

        return next.get(tx, target, indices);
    }
    
    public void remove(Transaction tx, GlobalIdentity id) throws CacheException {
        if (null == tx) {
            next.remove(tx, id);
            return;
        }
        Map cache = getTxCache(tx);

        CacheRowEntry row = (CacheRowEntry) cache.get(id);
        if (null == row) {
            return;
        }
        row.markRemoved();
    }
    
    public void release(GlobalIdentity id) throws CacheException {
        next.release(id);
    }

    public void clean() throws CacheException {
        next.clean();
    }
    
    private Map getTxCache(Transaction tx) throws CacheException {
        Map cache;
        synchronized (txToCache) {
            cache = (Map) txToCache.get(tx);
            if (null == cache) {
                cache = new HashMap();
                txToCache.put(tx, cache);
                try {
                    tx.registerSynchronization(new TxCompletionListener(tx));
                } catch (Exception e) {
                    throw new CacheException(e);
                }
            }
        }
        return cache;
    }

    private class TxCompletionListener implements Synchronization {
        private final Transaction tx;
        
        private TxCompletionListener(Transaction tx) {
            this.tx = tx;
        }
        
        public void beforeCompletion() {
        }

        public void afterCompletion(int status) {
            Map cache;
            synchronized (txToCache) {
                cache = (Map) txToCache.remove(tx);
            }
            try {
                next.transactionEnd(tx, status, cache.values());
            } catch (CacheException e) {
                throw new RuntimeCacheException(e);
            }
        }
    }
}