/**
 *
 *  Copyright 2004 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.pkgenerator;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import javax.sql.DataSource;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.tranql.cache.CacheRow;
import org.tranql.cache.DuplicateIdentityException;
import org.tranql.cache.InTxCache;
import org.tranql.identity.GlobalIdentity;
import org.tranql.sql.jdbc.JDBCUtil;

/**
 *
 *
 *
 * @version $Revision: 253 $ $Date: 2005-09-13 09:03:37 -0400 (Tue, 13 Sep 2005) $
 */
public class SequenceTablePrimaryKeyGenerator implements PrimaryKeyGenerator {
    private static final Log log = LogFactory.getLog(SequenceTablePrimaryKeyGenerator.class);
    
    private final TransactionManager tm;
    private final DataSource ds;
    private final String tableName;
    private final String sequenceName;
    private final int batchSize;
    private final String updateSequence;
    private final String selectSequence;
    
    private final Integer[] cache;
    private int cacheIdx;
    
    public SequenceTablePrimaryKeyGenerator(TransactionManager tm, DataSource ds, String tableName, String sequenceName, int batchSize) {
        this.tm = tm;
        this.ds = ds;
        this.tableName = tableName;
        this.sequenceName = sequenceName;
        this.batchSize = batchSize;

        cache = new Integer[batchSize];
        cacheIdx = 0;
        
        updateSequence = "UPDATE " + tableName + " SET value = value + " + batchSize +
            " WHERE name = '" + sequenceName + "'";
        selectSequence = "SELECT value FROM " + tableName + " WHERE name = '" + sequenceName + "'";
    }

    public void initSequenceTable() throws PrimaryKeyGeneratorException {
        Connection c = null;
        Statement s = null;
        ResultSet rs = null;
        try {
            c = ds.getConnection();
            DatabaseMetaData databaseMetaData = c.getMetaData();
            if ( false == databaseMetaData.supportsTransactionIsolationLevel(Connection.TRANSACTION_SERIALIZABLE) ) {
                log.warn("Database does not support isolation level TRANSACTION_SERIALIZABLE. " +
                        "Only one instance of SequenceTablePrimaryKey MUST be used.");
            }
            s = c.createStatement();
            s.executeUpdate("CREATE TABLE " + tableName + "(name VARCHAR(50), value INTEGER)");
        } catch (SQLException e) {
            ;
        } finally {
            JDBCUtil.close(c, s, rs);
        }

        try {
            c = ds.getConnection();
            s = c.createStatement();
            rs = s.executeQuery(selectSequence);
            if ( false == rs.next() ) {
                s.executeUpdate("INSERT INTO " + tableName + " (name, value) VALUES ('" + sequenceName + "', 0)");
            }
        } catch (SQLException e1) {
            throw new PrimaryKeyGeneratorException("Can not retrieve value from sequence table.", e1);
        } finally {
            JDBCUtil.close(c, s, rs);
        }

    }
    
    public synchronized Object getNextPrimaryKey(CacheRow cacheRow) throws PrimaryKeyGeneratorException {
        Object result = checkCache();
        if ( null != result ) {
            return result;
        }
        Connection c = null;
        Statement s = null;
        ResultSet rs = null;
        Transaction transaction = null;
        try {
            transaction = tm.suspend();
            
            tm.begin();
            
            c = ds.getConnection();
            DatabaseMetaData databaseMetaData = c.getMetaData();
            if ( databaseMetaData.supportsTransactionIsolationLevel(Connection.TRANSACTION_SERIALIZABLE) ) {
                c.setTransactionIsolation(Connection.TRANSACTION_SERIALIZABLE);
            }
            
            s = c.createStatement();
            rs = s.executeQuery(selectSequence);
            if ( false == rs.next() ) {
                throw new PrimaryKeyGeneratorException("Sequence " + sequenceName + " defined by sequence table " +
                        tableName + " is corrupted.");
            }
            fillCache(rs.getInt(1));
            int succUpdate = s.executeUpdate(updateSequence);
            
            tm.commit();

            if ( 1 != succUpdate ) {
                throw new PrimaryKeyGeneratorException("Sequence " + sequenceName + " defined by sequence table " +
                        tableName + " is corrupted.");
            }
            return checkCache();
        } catch (Exception e) {
            throw new PrimaryKeyGeneratorException("Unable to generateQuery new identity value", e);
        } finally {
            JDBCUtil.close(c, s, rs);
            try {
                if (null != transaction) {
                    tm.resume(transaction);
                }
            } catch (Exception e) {
                throw new PrimaryKeyGeneratorException("Unable to resume transaction", e);
            }
        }
    }

    public CacheRow updateCache(InTxCache cache, GlobalIdentity id, CacheRow cacheRow) throws DuplicateIdentityException {
        return id.getTable().addRow(cache, id, cacheRow);
    }

    private Integer checkCache() {
        if ( batchSize == cacheIdx ) {
            return null;
        }
        return cache[cacheIdx++];
    }
    
    private void fillCache(int minSequence) {
        for (int i = 0; i < cache.length; i++) {
            cache[i] = new Integer(minSequence + i);
        }
        cacheIdx = 0;
    }
}
