/**
 *
 *  Copyright 2004 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.ql;

import java.io.Serializable;
import java.util.List;

import org.tranql.schema.Attribute;

/**
 * This node represents an aggregate function call.  This is used within a
 * select clause.  This node expects one child.
 *
 * @version $Revision: 410 $ $Date: 2006-10-21 20:15:17 -0400 (Sat, 21 Oct 2006) $
 */
public class AggregateFunction extends AbstractNode {
    public static final FunctionType AVG = new FunctionTypeImpl("AVG");
    public static final FunctionType MAX = new FunctionTypeImpl("MAX");
    public static final FunctionType MIN = new FunctionTypeImpl("MIN");
    public static final FunctionType SUM = new FunctionTypeImpl("SUM");

    public interface FunctionType extends Serializable {
        String getName();

        void checkType(Node node);
        
        Class getType(Node node);
    }
    
    public static class FunctionTypeImpl implements FunctionType {
        protected final String name;
        
        protected FunctionTypeImpl(String name) {
            this.name = name;
        }
        
        public void checkType(Node node){
            if (false == node instanceof Path) {
                throw new MalformedNodeException(name + " expect a cmp-field path. Specified path is " + node);
            }
            
            Path path = (Path) node;
            List thePath = path.getPath();
            Object lastElement = thePath.get(thePath.size() - 1);
            if (false == lastElement instanceof Attribute) {
                throw new MalformedNodeException(name + " expect a cmp-field path. Specifid path is " + path);
            }
        }
        
        public Class getType(Node node) {
            Path path = (Path) node;
            List thePath = path.getPath();
            Attribute attribute = (Attribute) thePath.get(thePath.size() - 1);
            return attribute.getType();
        }

        public String getName() {
            return name;
        }
    }
    
    private final FunctionType type;
    private final boolean distinct;
    
    public AggregateFunction(FunctionType type, boolean distinct) {
        this.type = type;
        this.distinct = distinct;
    }

    public String getFunction() {
        return type.getName();
    }

    public FunctionType getFunctionType() {
        return type;
    }
    
    public Class getType() {
        return type.getType(this.getChild());
    }

    public boolean isDistinct() {
        return distinct;
    }

    public Path getPath() {
        return (Path) getChild();
    }

    public Node addChild(Node child) {
        type.checkType(child);
        return super.addChild(child);
    }
    
    public Object visit(QueryVisitor visitor, Object param) throws QueryException {
        return visitor.visit(this, param);
    }
    
}
