/**
 *
 *  Copyright 2004 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.query;

import org.tranql.ql.AggregateFunction;
import org.tranql.ql.Assignment;
import org.tranql.ql.AttributeReference;
import org.tranql.ql.Between;
import org.tranql.ql.BinaryOperation;
import org.tranql.ql.BooleanLiteral;
import org.tranql.ql.ConditionalAssignment;
import org.tranql.ql.CountAggregateFunction;
import org.tranql.ql.Delete;
import org.tranql.ql.DerivedTable;
import org.tranql.ql.Empty;
import org.tranql.ql.EntityReference;
import org.tranql.ql.Exists;
import org.tranql.ql.FKAttributeReference;
import org.tranql.ql.From;
import org.tranql.ql.Function;
import org.tranql.ql.In;
import org.tranql.ql.Insert;
import org.tranql.ql.IsNull;
import org.tranql.ql.Join;
import org.tranql.ql.Like;
import org.tranql.ql.MemberOf;
import org.tranql.ql.Node;
import org.tranql.ql.Not;
import org.tranql.ql.NumericLiteral;
import org.tranql.ql.OrderBy;
import org.tranql.ql.ParameterReference;
import org.tranql.ql.Path;
import org.tranql.ql.Query;
import org.tranql.ql.QueryException;
import org.tranql.ql.QuerySource;
import org.tranql.ql.QueryVisitor;
import org.tranql.ql.RowConstructor;
import org.tranql.ql.Select;
import org.tranql.ql.SetList;
import org.tranql.ql.StringLiteral;
import org.tranql.ql.SubQuery;
import org.tranql.ql.SubQuerySource;
import org.tranql.ql.TableConstructor;
import org.tranql.ql.UnaryOperator;
import org.tranql.ql.Update;
import org.tranql.ql.Where;

/**
 *
 *
 * @version $Revision: 410 $ $Date: 2006-10-21 20:15:17 -0400 (Sat, 21 Oct 2006) $
 */
public abstract class QueryCloner implements CommandTransform, QueryVisitor {

    protected void visitChildren(Node to, Node from, Object param) throws QueryException {
        for (Node fromChild = from.getChild(); fromChild != null; fromChild = fromChild.getSibling()) {
            Node toChild = (Node) fromChild.visit(this, param);
            if (toChild != null) {
                to.addChild(toChild);
            }
        }
    }

    public Object visit(SetList node, Object param) throws QueryException {
        SetList to = new SetList();
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(Assignment node, Object param) throws QueryException {
        Assignment to = new Assignment();
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(ConditionalAssignment node, Object param) throws QueryException {
        ConditionalAssignment to = new ConditionalAssignment();
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(From node, Object param) throws QueryException {
        From to = new From();
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(Select node, Object param) throws QueryException {
        Select to = new Select(node.isDistinct());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(Join node, Object param) throws QueryException {
        Join to = new Join(node.getType());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(Where node, Object param) throws QueryException {
        Where to = new Where();
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(OrderBy node, Object param) throws QueryException {
        OrderBy to = new OrderBy(node.isAsc());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(TableConstructor node, Object param) throws QueryException {
        TableConstructor to = new TableConstructor();
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(RowConstructor node, Object param) throws QueryException {
        RowConstructor to = new RowConstructor();
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(Between node, Object param) throws QueryException {
        Between to = new Between(node.isNot());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(In node, Object param) throws QueryException {
        In to = new In(node.isNot());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(Like node, Object param) throws QueryException {
        Like to = new Like(node.isNot());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(Empty node, Object param) throws QueryException {
        Empty to = new Empty(node.isNot());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(Exists node, Object param) throws QueryException {
        Exists to = new Exists(node.isNot());
        visitChildren(to, node, param);
        return to;
    }
    
    public Object visit(IsNull node, Object param) throws QueryException {
        IsNull to = new IsNull(node.isNot());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(MemberOf node, Object param) throws QueryException {
        MemberOf to = new MemberOf(node.isNot());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(SubQuery node, Object param) throws QueryException {
        SubQuery to = new SubQuery();
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(SubQuerySource node, Object param) throws QueryException {
        SubQuerySource to = new SubQuerySource(node.getSubQuery(), node.getAlias());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(NumericLiteral node, Object param) throws QueryException {
        NumericLiteral to = new NumericLiteral(node.getValue(), node.getText());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(BooleanLiteral node, Object param) throws QueryException {
        BooleanLiteral to = new BooleanLiteral(node.getValue());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(StringLiteral node, Object param) throws QueryException {
        StringLiteral to = new StringLiteral(node.getValue());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(ParameterReference node, Object param) throws QueryException {
        ParameterReference to = new ParameterReference(node.getParam());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(BinaryOperation node, Object param) throws QueryException {
        BinaryOperation to = new BinaryOperation(node.getOperation());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(Not node, Object param) throws QueryException {
        Not to = new Not();
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(Function node, Object param) throws QueryException {
        Function to = new Function(node.getFunctionType());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(AggregateFunction node, Object param) throws QueryException {
        AggregateFunction to = new AggregateFunction(node.getFunctionType(), node.isDistinct());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(CountAggregateFunction node, Object param) throws QueryException {
        CountAggregateFunction to = new CountAggregateFunction(node.isDistinct());
        visitChildren(to, node, param);
        return to;
    }
    
    public Object visit(Insert node, Object param) throws QueryException {
        Insert to = new Insert(node.getEntity());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(Update node, Object param) throws QueryException {
        Update to = new Update(node.getEntity());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(Delete node, Object param) throws QueryException {
        Delete to = new Delete(node.getEntity());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(Query node, Object param) throws QueryException {
        Query to = new Query(node);
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(QuerySource node, Object param) throws QueryException {
        QuerySource to = new QuerySource(node.getEntity(), node.getAlias(), node.isForUpdate());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(DerivedTable node, Object param) throws QueryException {
        DerivedTable to = new DerivedTable(node.getPath(), node.getAlias());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(EntityReference node, Object param) throws QueryException {
        EntityReference to = new EntityReference(node.getEntity());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(AttributeReference node, Object param) throws QueryException {
        AttributeReference to = new AttributeReference(node.getSource(), node.getAttribute());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(FKAttributeReference node, Object param) throws QueryException {
        FKAttributeReference to = new FKAttributeReference(node.getSource(), node.getFKAttribute());
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(Path node, Object param) throws QueryException {
        Path to = new Path(node.getRoot(), node.getPath());
        visitChildren(to, node, param);
        return to;
    }
    
    public Object visit(UnaryOperator node, Object param) throws QueryException {
        UnaryOperator to = new UnaryOperator(node.getType());
        visitChildren(to, node, param);
        return to;
    }
    
    public QueryCommand transform(QueryCommand source) throws QueryException {
        throw new UnsupportedOperationException();
    }

    public UpdateCommand transform(UpdateCommand source) throws QueryException {
        throw new UnsupportedOperationException();
    }

}
