/**
 *
 *  Copyright 2004 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.query;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.tranql.ql.AliasedEntity;
import org.tranql.ql.AttributeReference;
import org.tranql.ql.Delete;
import org.tranql.ql.DerivedTable;
import org.tranql.ql.EntityReference;
import org.tranql.ql.FKAttributeReference;
import org.tranql.ql.Insert;
import org.tranql.ql.Path;
import org.tranql.ql.Query;
import org.tranql.ql.QueryBinding;
import org.tranql.ql.QueryException;
import org.tranql.ql.QuerySource;
import org.tranql.ql.SubQuery;
import org.tranql.ql.SubQuerySource;
import org.tranql.ql.Update;
import org.tranql.ql.queryvisitor.ExecutorVisitor;
import org.tranql.ql.queryvisitor.TraverseAllNodeExecutor;
import org.tranql.schema.AssociationEnd;
import org.tranql.schema.Attribute;
import org.tranql.schema.Entity;
import org.tranql.schema.FKAttribute;
import org.tranql.schema.Schema;

/**
 *
 *
 * @version $Revision: 323 $ $Date: 2006-02-19 01:29:24 -0500 (Sun, 19 Feb 2006) $
 */
public class SchemaMapper extends QueryCloner {
    private final Schema target;

    public SchemaMapper(Schema target) {
        this.target = target;
    }

    public QueryCommand transform(QueryCommand source) throws QueryException {
        Context ctx = new Context();
        Query targetQuery = (Query) (source.getQuery().visit(this, ctx));
        return target.getCommandFactory().createQuery(targetQuery);
    }

    public UpdateCommand transform(UpdateCommand source) throws QueryException {
        Context ctx = new Context();
        Query targetQuery = (Query) (source.getQuery().visit(this, ctx));
        return target.getCommandFactory().createUpdate(targetQuery);
    }
    
    private Entity map(Entity from) throws QueryException {
        Entity entity = target.getEntity(from.getName());
        if (entity == null) {
            throw new QueryException("No target Entity defined for " + from.getName());
        }
        return entity;
    }

    private Attribute map(Entity source, Attribute from) throws QueryException {
        Attribute attr = source.getAttribute(from.getName());
        if (attr == null) {
            throw new QueryException("Target " + source.getName() + " does not have attribute " + from.getName());
        }
        return attr;
    }

    private Attribute map(AliasedEntity source, Attribute from) throws QueryException {
        return map(source.getEntity(), from);
    }

    private FKAttribute map(Entity source, FKAttribute from) throws QueryException {
        List ends = source.getAssociationEnds();
        for (Iterator iter = ends.iterator(); iter.hasNext();) {
            AssociationEnd end = (AssociationEnd) iter.next();

            // FK do not exist for these relationships.
            if (end.isManyToMany() || end.isOneToMany() || end.isOnPKSide()) {
                continue;
            }
            
            LinkedHashMap mapping =  end.getAssociation().getJoinDefinition().getPKToFKMapping();
            for (Iterator iterator = mapping.entrySet().iterator(); iterator.hasNext();) {
                Map.Entry entry = (Map.Entry) iterator.next();
                FKAttribute att = (FKAttribute)entry.getValue();
                if (att.getName().equals(from.getName())) {
                    return att; 
                }
            }
        }
        throw new QueryException("Target " + source.getName() + " does not have foreign key attribute " + from.getName());
    }

    private FKAttribute map(AliasedEntity source, FKAttribute from) throws QueryException {
        return map(source.getEntity(), from);
    }

    private AssociationEnd map(Entity source, AssociationEnd from) throws QueryException {
        AssociationEnd associationEnd = source.getAssociationEnd(from.getName());
        if (associationEnd == null) {
            throw new QueryException("Target " + source.getName() + " does not have association end " + from.getName());
        }
        return associationEnd;
    }

    private AssociationEnd map(AliasedEntity source, AssociationEnd from) throws QueryException {
        return map(source.getEntity(), from);
    }

    public static final class Context {
        private final Map aliases = new HashMap();
    }

    public Object visit(Query node, Object param) throws QueryException {
        QueryBinding[] paramBindings = node.getParamTransforms();
        for (int i = 0; i < paramBindings.length; i++) {
            QueryBinding binding = paramBindings[i];
            if (binding.isAttributeBinding()) {
                binding.setAttribute(map(map(binding.getEntity()), binding.getAttribute()));
            } else if (binding.isFKAttributeBinding()) {
                binding.setFKAttribute(map(map(binding.getEntity()), binding.getFKAttribute()));
            }
        }
        
        QueryBinding[] resultBindings = node.getResultTransforms();
        for (int i = 0; i < resultBindings.length; i++) {
            QueryBinding binding = resultBindings[i];
            if (binding.isAttributeBinding()) {
                binding.setAttribute(map(map(binding.getEntity()), binding.getAttribute()));
            } else if (binding.isFKAttributeBinding()) {
                binding.setFKAttribute(map(map(binding.getEntity()), binding.getFKAttribute()));
            }
        }
        
        Query to = new Query(paramBindings, resultBindings);
        to.setResultAccessors(node.getResultAccessors());
        
        AliasDefiner aliasDefiner = new AliasDefiner();
        node.visit(aliasDefiner, param);

        visitChildren(to, node, param);
        return to;
    }

    public Object visit(Insert node, Object param) throws QueryException {
        Context ctx = (Context) param;
        Insert to = new Insert(map(node.getEntity()));
        ctx.aliases.put(to.getAlias(), to);
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(Update node, Object param) throws QueryException {
        Context ctx = (Context) param;
        Update to = new Update(map(node.getEntity()));
        ctx.aliases.put(to.getAlias(), to);
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(Delete node, Object param) throws QueryException {
        Context ctx = (Context) param;
        Delete to = new Delete(map(node.getEntity()));
        ctx.aliases.put(to.getAlias(), to);
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(QuerySource node, Object param) throws QueryException {
        Context ctx = (Context) param;
        QuerySource to = new QuerySource(map(node.getEntity()), node.getAlias(), node.isForUpdate());
        ctx.aliases.put(to.getAlias(), to);
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(SubQuerySource node, Object param) throws QueryException {
        Context ctx = (Context) param;
        SubQuery subQuery = (SubQuery) node.getSubQuery().visit(this, param);
        SubQuerySource to = new SubQuerySource(subQuery, node.getAlias());
        ctx.aliases.put(to.getAlias(), to);
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(DerivedTable node, Object param) throws QueryException {
        Context ctx = (Context) param;
        Path path = node.getPath();
        AliasedEntity newAlias = (AliasedEntity) ctx.aliases.get(path.getRoot().getAlias());
        Entity newEntity = newAlias.getEntity();
        List newPath = new ArrayList();
        for (int i = 0; i < path.getPath().size(); i++) {
            AssociationEnd associationEnd = (AssociationEnd) path.getPath().get(i); 
            newPath.add(map(newEntity, associationEnd));
            newEntity = map(associationEnd.getEntity());
        }
        DerivedTable to = new DerivedTable(new Path(newAlias, newPath), node.getAlias());
        ctx.aliases.put(to.getAlias(), to);
        visitChildren(to, node, param);
        return to;
    }
    
    public Object visit(Path node, Object param) throws QueryException {
        Context ctx = (Context) param;
        AliasedEntity newAlias = (AliasedEntity) ctx.aliases.get(node.getRoot().getAlias());
        
        Entity newEntity = newAlias.getEntity();
        List newPath = new ArrayList();
        for (int i = 0; i < node.getPath().size(); i++) {
            Object opaque = node.getPath().get(i);
            if (opaque instanceof Attribute) {
                newPath.add(map(newEntity, (Attribute) opaque));
            } else if (opaque instanceof AssociationEnd) {
                AssociationEnd associationEnd = (AssociationEnd) opaque;
                newPath.add(map(newAlias, associationEnd));
                newEntity = map(associationEnd.getEntity());
            } else {
                throw new AssertionError();
            }
        }
        
        Path to = new Path(newAlias, newPath);
        visitChildren(to, node, param);
        return to;
    }
    
    public Object visit(AttributeReference node, Object param) throws QueryException {
        Context ctx = (Context) param;
        AliasedEntity target = (AliasedEntity) ctx.aliases.get(node.getSource().getAlias());
        if (target == null) {
            throw new QueryException("Alias is not defined: " + node.getSource().getAlias());
        }
        AttributeReference to = new AttributeReference(target, map(target, node.getAttribute()));
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(FKAttributeReference node, Object param) throws QueryException {
        Context ctx = (Context) param;
        AliasedEntity target = (AliasedEntity) ctx.aliases.get(node.getSource().getAlias());
        if (target == null) {
            throw new QueryException("Alias is not defined: " + node.getSource().getAlias());
        }
        FKAttributeReference to = new FKAttributeReference(target, map(target, node.getFKAttribute()));
        visitChildren(to, node, param);
        return to;
    }

    public Object visit(EntityReference node, Object param) throws QueryException {
        Context ctx = (Context) param;
        AliasedEntity target = (AliasedEntity) ctx.aliases.get(node.getEntity().getAlias());
        if (target == null) {
            throw new QueryException("Alias is not defined: " + node.getEntity().getAlias());
        }
        EntityReference to = new EntityReference(target);
        visitChildren(to, node, param);
        return to;
    }
    
    private class AliasDefiner extends ExecutorVisitor {
        public AliasDefiner() {
            super(new TraverseAllNodeExecutor());
            ((TraverseAllNodeExecutor) executor).setQueryVisitor(this);
        }
        
        public Object visit(Insert node, Object param) throws QueryException {
            return SchemaMapper.this.visit(node, param);
        }

        public Object visit(Update node, Object param) throws QueryException {
            return SchemaMapper.this.visit(node, param);
        }

        public Object visit(Delete node, Object param) throws QueryException {
            return SchemaMapper.this.visit(node, param);
        }

        public Object visit(QuerySource node, Object param) throws QueryException {
            return SchemaMapper.this.visit(node, param);
        }

        public Object visit(SubQuerySource node, Object param) throws QueryException {
            super.visit(node.getSubQuery(), param);
            return SchemaMapper.this.visit(node, param);
        }

        public Object visit(DerivedTable node, Object param) throws QueryException {
            return SchemaMapper.this.visit(node, param);
        }

    }
}
