/**
 *
 *  Copyright 2004 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.sql;

import java.io.StringReader;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.tranql.builder.IdentityDefinerBuilder;
import org.tranql.cache.GlobalSchema;
import org.tranql.cache.NotNullAccessor;
import org.tranql.ejb.CMPField;
import org.tranql.ejb.EJB;
import org.tranql.ejb.EJBLocalObjectAsIdTransform;
import org.tranql.ejb.EJBObjectAsIdTransform;
import org.tranql.ejb.EJBProxyFactory;
import org.tranql.ejb.EJBQLQuery;
import org.tranql.ejb.EJBSchema;
import org.tranql.ejb.FinderEJBQLQuery;
import org.tranql.ejb.SelectEJBQLQuery;
import org.tranql.ejb.parser.EJBQL21Compiler;
import org.tranql.ejb.parser.EJBQL21Parser;
import org.tranql.ejb.parser.EJBQLLexer;
import org.tranql.field.DomainIdentityAccessor;
import org.tranql.field.FieldAccessor;
import org.tranql.field.FieldTransform;
import org.tranql.field.ReferenceAccessor;
import org.tranql.identity.IdentityDefiner;
import org.tranql.identity.IdentityTransform;
import org.tranql.ql.Query;
import org.tranql.ql.QueryBinding;
import org.tranql.ql.QueryBindingImpl;
import org.tranql.ql.QueryException;
import org.tranql.query.QueryCommand;
import org.tranql.query.SchemaMapper;

import antlr.Token;
import antlr.collections.AST;


/**
 * 
 * 
 * @version $Revision: 340 $ $Date: 2006-06-03 18:13:16 -0400 (Sat, 03 Jun 2006) $
 */
public class EJBQLToPhysicalQuery implements EJBQLQueryToSQLCommandCompiler {
    private static final Log log = LogFactory.getLog(EJBQLToPhysicalQuery.class);
    
    private final EJBSchema ejbSchema;
    private final SchemaMapper mapper;

    private final IdentityDefinerBuilder identityDefinerBuilder;
    private final EJBQLCompilerFactory.EJBQLCompiler compiler;
    
    public EJBQLToPhysicalQuery(EJBSchema ejbSchema, SQLSchema sqlSchema, GlobalSchema globalSchema) {
        this.ejbSchema = ejbSchema;
        this.compiler = sqlSchema.getEJBQLCompilerFactory().factoryEJBQLCompiler(ejbSchema, sqlSchema, globalSchema);
        
        mapper = new SchemaMapper(sqlSchema);
        
        identityDefinerBuilder = new IdentityDefinerBuilder(ejbSchema, globalSchema);
    }
    
    public Map buildFinders(EJB ejb) throws QueryException {
        Map findersMap = new HashMap();
        
        Collection finders = ejb.getFinders();
        for (Iterator iter = finders.iterator(); iter.hasNext();) {
            FinderEJBQLQuery finder = (FinderEJBQLQuery) iter.next();
            QueryCommand[] commands;
            try {
                commands = buildFinder(ejb, finder);
            } catch (QueryException e) {
                throw new QueryException("EJB [" + ejb + "]", e);
            }
            log.debug("[" + finder + "] compiled to [" + commands[0] + "]");
            findersMap.put(finder, commands);
        }
        
        return findersMap;
    }
    
    public Map buildSelects(EJB ejb) throws QueryException {
        Map selectsMap = new HashMap();
        
        Collection selects = ejb.getSelects();
        for (Iterator iter = selects.iterator(); iter.hasNext();) {
            SelectEJBQLQuery select = (SelectEJBQLQuery) iter.next();
            QueryCommand command;
            try {
                command = buildSelect(select);
            } catch (QueryException e) {
                throw new QueryException("EJB [" + ejb + "]", e);
            }
            log.debug("[" + select + "] compiled to [" + command + "]");
            selectsMap.put(select, command);
        }
        
        return selectsMap;
    }

    public QueryCommand buildQueryCommand(SelectEJBQLQuery query) throws QueryException {
        QueryCommand command = buildSelect(query);
        log.debug("[" + query + "] compiled to [" + command + "]");
        return command;
    }

    private QueryCommand[] buildFinder(EJB ejb, FinderEJBQLQuery finder) throws QueryException {
        QueryCommand localCommand;
        QueryCommand remoteCommand;
        try {
            Query query = getEJBQLQuery(finder);

            ReturnedValueTypeDetectorVisitor typeDetector = new ReturnedValueTypeDetectorVisitor();
            query = (Query) typeDetector.visit(query, null);
            
            ReturnedValueTypeDetectorVisitor.ReturnedValueTypeInfo info = typeDetector.getInfo();
            if (ejb != info.getEntity()) {
                throw new QueryException("Finder does not return the right abstract schema type.");
            }
            
            localCommand = buildView(ejb, query, true);
            remoteCommand = buildView(ejb, query, false);
        } catch (QueryException e) {
            throw new QueryException("Finder [" + finder + "]", e);
        }
        return new QueryCommand[]{localCommand, remoteCommand};
    }
    
    private QueryCommand buildSelect(SelectEJBQLQuery select) throws QueryException {
        QueryCommand command;
        Class returnedType;
        try {
            Query query = getEJBQLQuery(select);

            ReturnedValueTypeDetectorVisitor typeDetector = new ReturnedValueTypeDetectorVisitor();
            query = (Query) typeDetector.visit(query, null);

            ReturnedValueTypeDetectorVisitor.ReturnedValueTypeInfo info = typeDetector.getInfo();
            if (null != info.getEntity()) {
                EJB ejb = (EJB) info.getEntity();
                select.setSelectedEJB(ejb);
                return buildView(ejb, query, select.isLocal());
            }

            if (null != info.getAttribute()) {
                CMPField field = (CMPField) info.getAttribute();
                returnedType = field.getType();
            } else {
                returnedType = info.getAggregatedResultType();
            }

            command = expandToSQLQuery(query);
        } catch (QueryException e) {
            throw new QueryException("Select [" + select + "]", e);
        }
        
        command.getQuery().setResultAccessors(new FieldTransform[] {new FieldAccessor(0, returnedType)});
        
        command = mapper.transform(command);

        return command;
    }
    
    private QueryCommand buildView(EJB ejb, Query query, boolean isLocal) throws QueryException {
        IdentityDefiner identityDefiner = identityDefinerBuilder.getIdentityDefiner(ejb, 0);
        IdentityTransform identityTransform = identityDefinerBuilder.getPrimaryKeyTransform(ejb);
        EJBProxyFactory proxyFactory = ejb.getProxyFactory();

        FieldTransform nullTester = new ReferenceAccessor(identityDefiner);
        FieldTransform view = new DomainIdentityAccessor(nullTester, identityTransform);

        if (isLocal) {
            view = new EJBLocalObjectAsIdTransform(view, proxyFactory, ejb.getPrimaryKeyClass());
        } else {
            view = new EJBObjectAsIdTransform(view, proxyFactory, ejb.getPrimaryKeyClass());
        }
        view = new NotNullAccessor(nullTester, view);

        QueryCommand command = expandToSQLQuery(query);
        command.getQuery().setResultAccessors(new FieldTransform[] {view});
        
        command = mapper.transform(command);

        return command;
    }
    
    private Query getEJBQLQuery(EJBQLQuery ejbQL) throws QueryException {
        Class[] paramTypes = ejbQL.getParameterTypes();
        QueryBinding[] paramTransforms = new QueryBinding[paramTypes.length];
        for (int i = 0; i < paramTransforms.length; i++) {
            paramTransforms[i] = new QueryBindingImpl(i, paramTypes[i]);
        }
        
        Query query;
        try {
            EJBQLLexer lexer = new EJBQLLexer(new StringReader(ejbQL.getEjbQL()));
            EJBQL21Parser parser = new EJBQL21Parser(lexer);
            parser.ejbql();

            if (parser.hadErrors()) {
                // TODO extract the errors.
                throw new QueryException("Error while parsing.");
            }
            
            Token token = lexer.nextToken();
            if (token.getType() != Token.EOF_TYPE) {
                // TODO extract the remaining tokens.
                throw new QueryException("Tokens left on the stream.");
            }

            AST ast = parser.getAST();

            EJBQL21Compiler walker = new EJBQL21Compiler(paramTransforms, ejbSchema);
            query = walker.ejbql(ast);
        } catch (Exception e) {
            throw new QueryException("Parser, lexer or Tree-walker exception", e);
        }
        
        return query;
    }

    private QueryCommand expandToSQLQuery(Query query) throws QueryException {
        QueryCommand command = ejbSchema.getCommandFactory().createQuery(query); 
        return compiler.transform(command);
    }
    
}
