/**
 *
 *  Copyright 2004 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.sql;

import java.util.List;

import org.tranql.ejb.EJB;
import org.tranql.ql.AggregateFunction;
import org.tranql.ql.EntityReference;
import org.tranql.ql.Node;
import org.tranql.ql.Path;
import org.tranql.ql.Query;
import org.tranql.ql.QueryBinding;
import org.tranql.ql.QueryBindingImpl;
import org.tranql.ql.QueryException;
import org.tranql.ql.Select;
import org.tranql.ql.queryvisitor.ExecutorVisitor;
import org.tranql.ql.queryvisitor.TraverseAllNodeExecutor;
import org.tranql.query.QueryCloner;
import org.tranql.schema.AssociationEnd;
import org.tranql.schema.Attribute;
import org.tranql.schema.Entity;

/**
 * 
 * 
 * @version $Revision: 173 $ $Date: 2005-02-20 10:42:56 -0500 (Sun, 20 Feb 2005) $
 */
class ReturnedValueTypeDetectorVisitor extends QueryCloner {
    
    public interface ReturnedValueTypeInfo {
        Entity getEntity();
        Attribute getAttribute();
        Entity getAttributeEntity();
        Class getAggregatedResultType();
    }

    private static final InfoNotInitState NOT_INIT_STATE = new InfoNotInitState();
    
    private static final class InfoNotInitState implements ReturnedValueTypeInfo {
        public Entity getEntity() {
            throw new IllegalStateException("No Select has been visited.");
        }
        public Attribute getAttribute() {
            throw new IllegalStateException("No Select has been visited.");
        }
        public Entity getAttributeEntity() {
            throw new IllegalStateException("No Select has been visited.");
        }
        public Class getAggregatedResultType() {
            throw new IllegalStateException("No Select has been visited.");
        }
    }

    private static final class InfoInitState implements ReturnedValueTypeInfo {
        private final Entity entity;
        private final Attribute attribute;
        private final Entity attributeEntity;
        private final Class aggregatedResultType;
        private InfoInitState(Entity entity, Attribute attribute, Entity attributeEntity, Class aggregatedResultType) {
            this.entity = entity;
            this.attribute = attribute;
            this.attributeEntity = attributeEntity;
            this.aggregatedResultType = aggregatedResultType;
        }
        public Entity getEntity() {
            return entity;
        }
        public Attribute getAttribute() {
            return attribute;
        }
        public Entity getAttributeEntity() {
            return attributeEntity;
        }
        public Class getAggregatedResultType() {
            return aggregatedResultType;
        }
    }

    private ReturnedValueTypeInfo info = NOT_INIT_STATE;
    
    public ReturnedValueTypeInfo getInfo() {
        return info;
    }
    
    public Object visit(Query node, Object param) throws QueryException {
        RetrieveInfo retrieveInfo = new RetrieveInfo();
        retrieveInfo.visit(node, param);
        
        QueryBinding[] resultTransforms;
        if (null != info.getEntity()) {
            EJB ejb = (EJB) info.getEntity();
            List pkFields = ejb.getPrimaryKeyFields();
            resultTransforms = new QueryBinding[pkFields.size()];
            for (int i = 0; i < resultTransforms.length; i++) {
                Attribute attribute = (Attribute) pkFields.get(i);
                resultTransforms[i] = new QueryBindingImpl(i, ejb, attribute);
            }
        } else if (null != info.getAttribute()) {
            Entity entity = info.getAttributeEntity();
            Attribute attribute = info.getAttribute();
            resultTransforms = new QueryBinding[] {new QueryBindingImpl(0, entity, attribute)};
        } else {
            resultTransforms = new QueryBinding[] {new QueryBindingImpl(0, info.getAggregatedResultType())};
        }

        Query to = new Query(node.getParamTransforms(), resultTransforms);
        visitChildren(to, node, param);
        return to;
    }
    
    private class RetrieveInfo extends ExecutorVisitor {
        private RetrieveInfo() {
            super(new TraverseAllNodeExecutor());
            TraverseAllNodeExecutor allNodeExecutor = (TraverseAllNodeExecutor) executor;
            allNodeExecutor.setQueryVisitor(this);
        }
        
        public Object visit(Select node, Object param) throws QueryException {
            Node selected = node.getChild();
            if (selected instanceof Path) {
                Path path = (Path) selected;
                List thePath = path.getPath();
                Object lastElement = thePath.get(thePath.size() - 1);
                if (lastElement instanceof AssociationEnd) {
                    AssociationEnd end = (AssociationEnd) lastElement;
                    info = new InfoInitState(end.getEntity(), null, null, null);
                } else {
                    Entity attributeEntity;
                    Attribute attribute = (Attribute) lastElement;
                    if (1 == thePath.size()) {
                        attributeEntity = path.getRoot().getEntity();
                    } else {
                        AssociationEnd end = (AssociationEnd) thePath.get(thePath.size() - 2);
                        attributeEntity = end.getEntity();
                    }
                    info = new InfoInitState(null, attribute, attributeEntity, null);
                }
            } else if (selected instanceof EntityReference) {
                EntityReference reference = (EntityReference) selected;
                info = new InfoInitState(reference.getEntity().getEntity(), null, null, null);
            } else if (selected instanceof AggregateFunction) {
                AggregateFunction function = (AggregateFunction) selected;
                info = new InfoInitState(null, null, null, function.getType());
            }
            
            return null;
        }
    }
    
}
