/**
 *
 *  Copyright 2005 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.sql.prefetch;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;

import org.tranql.ql.QueryBinding;
import org.tranql.ql.QueryBindingImpl;
import org.tranql.schema.AssociationEnd;
import org.tranql.schema.Attribute;
import org.tranql.schema.Entity;
import org.tranql.schema.Schema;
import org.tranql.sql.jdbc.binding.BindingFactory;
import org.tranql.sql.jdbc.binding.BindingFactory.BindingInfo;
import org.tranql.sql.prefetch.PrefetchGroup.AssociationEndInfo;

/**
 *
 *
 * @version $Revision: 468 $ $Date: 2006-12-13 16:58:42 -0500 (Wed, 13 Dec 2006) $
 */
public class PrefetchGroupDictionary {
    private static final String DEFAULT_GROUP = "DEFAULT GROUP";
    private static final Attribute[] EMPTY_ATTRIBUTES = new Attribute[0];
    private static final AssociationEndDesc[] EMPTY_ENDS = new AssociationEndDesc[0];
    
    private final Schema schema;
    private final Map groupKeyToGroupInfo = new HashMap();

    public PrefetchGroupDictionary(Schema schema) {
        this.schema = schema;
    }
    
    public void addPrefetchGroup(String groupName, String entityName, String[] attributeNames, AssociationEndDesc[] ends) {
        Entity entity = schema.getEntity(entityName);
        EntityGroupKey entityGroupKey = new EntityGroupKey(groupName, entity);
        if (null != groupKeyToGroupInfo.get(entityGroupKey)) {
            throw new IllegalArgumentException("Group [" + groupName + "] already defined by table [" + entity + "].");
        }
        Attribute[] attributes = new Attribute[attributeNames.length];
        for (int i = 0; i < attributeNames.length; i++) {
            attributes[i] = entity.getAttribute(attributeNames[i]);
            if (null == attributes[i]) {
                throw new IllegalArgumentException("No column named [" + attributeNames[i] + "] defined by table [" + entity + "].");
            }
        }
        for (int i = 0; i < ends.length; i++) {
            ends[i].associationEnd = entity.getAssociationEnd(ends[i].end);
            if (null == ends[i].associationEnd) {
                throw new IllegalArgumentException("No end table named [" + ends[i].end + "] defined by table [" + entity + "].");
            }
        }
        
        groupKeyToGroupInfo.put(entityGroupKey, new PrefetchGroupInfo(groupName, entity, attributes, ends));
    }

    public PrefetchGroup getGroupByAttributeName(String entityName, String attributeName) {
        Entity entity = schema.getEntity(entityName);
        Attribute attribute = entity.getAttribute(attributeName);
        for (Iterator iter = groupKeyToGroupInfo.entrySet().iterator(); iter.hasNext();) {
            Map.Entry entry = (Map.Entry) iter.next();
            EntityGroupKey groupKey = (EntityGroupKey) entry.getKey();
            if (groupKey.entity != entity) {
                continue;
            }
            PrefetchGroupInfo groupInfo = (PrefetchGroupInfo) entry.getValue();
            if (groupInfo.defineAttribute(attribute)) {
                return snapshotPrefetchGroup(groupInfo, new HashSet());
            }
        }
        
        return null;
    }

    public PrefetchGroup getGroupByEndName(String entityName, String endName) {
        Entity entity = schema.getEntity(entityName);
        AssociationEnd end = entity.getAssociationEnd(endName);
        for (Iterator iter = groupKeyToGroupInfo.entrySet().iterator(); iter.hasNext();) {
            Map.Entry entry = (Map.Entry) iter.next();
            EntityGroupKey groupKey = (EntityGroupKey) entry.getKey();
            if (groupKey.entity != entity) {
                continue;
            }
            PrefetchGroupInfo groupInfo = (PrefetchGroupInfo) entry.getValue();
            if (groupInfo.defineAssociationEnd(end)) {
                return snapshotPrefetchGroup(groupInfo, new HashSet());
            }
        }
        
        return null;
    }

    public PrefetchGroup getPrefetchGroup(String groupName, String entityName) {
        Entity entity = schema.getEntity(entityName);
        PrefetchGroupInfo groupInfo = getPrefetchGroupInfo(groupName, entity);
        if (null == groupInfo) {
            throw new IllegalArgumentException("Group [" + groupName + "] is not defined by entity [" + entityName + "].");
        }
        return snapshotPrefetchGroup(groupInfo, new HashSet());
    }

    public PrefetchGroup getDefaultPrefetchGroup(String entityName) {
        Entity entity = schema.getEntity(entityName);
        List attributes = entity.getAttributes();
        List columns = new ArrayList();
        for (Iterator iter = attributes.iterator(); iter.hasNext();) {
            Attribute attribute = (Attribute) iter.next();
            QueryBinding binding = new QueryBindingImpl(0, entity, attribute);
            BindingInfo info = BindingFactory.getBindingInfo(binding);
            if (info.isLob()) {
                continue;
            }
            columns.add(attribute);
        }
        Attribute[] prefetched = (Attribute[]) columns.toArray(new Attribute[columns.size()]);
        PrefetchGroupInfo groupInfo = new PrefetchGroupInfo(DEFAULT_GROUP, entity, prefetched, EMPTY_ENDS);
        return snapshotPrefetchGroup(groupInfo, new HashSet());
    }

    private PrefetchGroupInfo getPrefetchGroupInfo(String groupName, Entity entity) {
        return (PrefetchGroupInfo) groupKeyToGroupInfo.get(new EntityGroupKey(groupName, entity));
    }

    private PrefetchGroup snapshotPrefetchGroup(PrefetchGroupInfo groupInfo, Set traversedEnds) {
        AssociationEndDesc[] ends = groupInfo.ends;
        Collection endInfoList = new ArrayList();
        for (int i = 0; i < ends.length; i++) {
            AssociationEnd end = ends[i].associationEnd;
            String groupName = ends[i].groupName;
            if (traversedEnds.contains(end)) {
                continue;
            }
            traversedEnds.add(end);
            traversedEnds.add(end.getAssociation().getOtherEnd(end));
            PrefetchGroupInfo childGroupInfo = getPrefetchGroupInfo(groupName, end.getEntity());
            if (null == childGroupInfo) {
                childGroupInfo = new PrefetchGroupInfo(DEFAULT_GROUP, end.getEntity(), EMPTY_ATTRIBUTES, EMPTY_ENDS);
            }
            PrefetchGroup childGroup = snapshotPrefetchGroup(childGroupInfo, traversedEnds);
            endInfoList.add(new PrefetchGroupImpl.AssociationEndInfoImpl(end, childGroup));
        }
        AssociationEndInfo[] endInfo = (AssociationEndInfo[]) endInfoList.toArray(new AssociationEndInfo[0]); 
        return new PrefetchGroupImpl(groupInfo, endInfo);
    }
    
    private class PrefetchGroupInfo {
        private final String name;
        private final Entity entity;
        private final Attribute[] attributes;
        private final AssociationEndDesc[] ends;

        private PrefetchGroupInfo(String name, Entity entity, Attribute[] columns, AssociationEndDesc[] ends) {
            this.name = name;
            this.entity = entity;
            this.attributes = columns;
            this.ends = ends;
        }
        
        public boolean defineAttribute(Attribute attribute) {
            for (int i = 0; i < attributes.length; i++) {
                if (attributes[i] == attribute) {
                    return true;
                }
            }
            return false;
        }

        public boolean defineAssociationEnd(AssociationEnd end) {
            for (int i = 0; i < ends.length; i++) {
                if (ends[i].associationEnd == end) {
                    return true;
                }
            }
            return false;
        }
    }
    
    public static class AssociationEndDesc {
        private final String end;
        private AssociationEnd associationEnd;
        private final String groupName;
        public AssociationEndDesc(String end, String groupName) {
            this.end = end;
            this.groupName = groupName;
        }
    }
    
    private static class PrefetchGroupImpl implements PrefetchGroup {
        private PrefetchGroup parent;
        private AssociationEndInfo parentAssociationEndInfo;
        private final String name;
        private final Entity entity;
        private final Attribute[] attributes;
        private final AssociationEndInfo[] endInfo;

        private PrefetchGroupImpl(PrefetchGroupInfo info, AssociationEndInfo[] endInfo) {
            this.name = info.name;
            this.entity = info.entity;
            this.attributes = info.attributes;
            this.endInfo = endInfo;
            for (int i = 0; i < endInfo.length; i++) {
                PrefetchGroupImpl groupImpl = (PrefetchGroupImpl) endInfo[i].getPrefetchGroup(); 
                groupImpl.parent = this;
                groupImpl.parentAssociationEndInfo = endInfo[i];
            }
        }

        public String getName() {
            return name;
        }

        public Entity getEntity() {
            return entity;
        }
        
        public Attribute[] getAttributes() {
            return attributes;
        }

        public AssociationEndInfo[] getEndTableInfo() {
            return endInfo;
        }
        
        public PrefetchGroupIterator getIterator() {
            return new PrefetchGroupIteratorImpl();
        }
        
        public PrefetchGroup getParentPrefetchGroup() {
            return parent;
        }

        public AssociationEndInfo getParentEndTableInfo() {
            return parentAssociationEndInfo;
        }
        
        private static class AssociationEndInfoImpl implements AssociationEndInfo {
            private final AssociationEnd end;
            private final PrefetchGroup prefetchGroup;
            
            private AssociationEndInfoImpl(AssociationEnd end, PrefetchGroup prefetchGroup) {
                this.end = end;
                this.prefetchGroup = prefetchGroup;
            }
            
            public AssociationEnd getAssociationEnd() {
                return end;
            }

            public PrefetchGroup getPrefetchGroup() {
                return prefetchGroup;
            }
        }
        
        private class PrefetchGroupIteratorImpl implements PrefetchGroupIterator {
            private PrefetchGroupIterator nestedGroupIterator;
            private int nestedGroupIndex;
            
            public PrefetchGroupIteratorImpl() {
                this.nestedGroupIndex = -1;
                if (0 == endInfo.length) {
                    nestedGroupIterator = new PrefetchGroupIterator() {
                        public boolean hasMorePrefetchGroups() {
                            return false;
                        }
                        public PrefetchGroup nextPrefetchGroup() {
                            throw new NoSuchElementException();
                        }
                    };
                } else {
                    AssociationEndInfo info = endInfo[0];
                    nestedGroupIterator = info.getPrefetchGroup().getIterator();
                }
            }

            public boolean hasMorePrefetchGroups() {
                if (-1 == nestedGroupIndex) {
                    return true;
                }
                if (nestedGroupIterator.hasMorePrefetchGroups()) {
                    return true;
                }
                if (endInfo.length <= nestedGroupIndex + 1) {
                    return false;
                }
                AssociationEndInfo info = endInfo[nestedGroupIndex + 1];
                PrefetchGroupIterator iterator = info.getPrefetchGroup().getIterator();
                return iterator.hasMorePrefetchGroups();
            }

            public PrefetchGroup nextPrefetchGroup() {
                if (-1 == nestedGroupIndex) {
                    nestedGroupIndex++;
                    return PrefetchGroupImpl.this;
                }
                if (nestedGroupIterator.hasMorePrefetchGroups()) {
                    return nestedGroupIterator.nextPrefetchGroup();
                }
                if (endInfo.length <= nestedGroupIndex + 1) {
                    throw new NoSuchElementException();
                }
                AssociationEndInfo info = endInfo[++nestedGroupIndex];
                nestedGroupIterator = info.getPrefetchGroup().getIterator();
                return nestedGroupIterator.nextPrefetchGroup();
            }
        }
    }

    private class EntityGroupKey {
        private final String groupName;
        private final Entity entity;
        
        private EntityGroupKey(String groupName, Entity entity) {
            this.groupName = groupName;
            this.entity = entity;
        }
        
        public int hashCode() {
            return groupName.hashCode() & entity.hashCode();
        }

        public boolean equals(Object obj) {
            if (false ==  obj instanceof EntityGroupKey) {
                return false;
            }
            EntityGroupKey other = (EntityGroupKey) obj;
            return  groupName.equals(other.groupName) && entity == other.entity;
        }
    }
}
