/**
 *
 *  Copyright 2005 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.sql.prefetch;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.List;

import org.tranql.builder.IdentityDefinerBuilder;
import org.tranql.cache.CacheRow;
import org.tranql.cache.EmptySlotLoader;
import org.tranql.cache.InTxCache;
import org.tranql.field.FieldAccessor;
import org.tranql.field.FieldTransform;
import org.tranql.field.Row;
import org.tranql.identity.GlobalIdentity;
import org.tranql.identity.IdentityDefiner;
import org.tranql.ql.QueryException;
import org.tranql.query.QueryCommand;
import org.tranql.query.ResultHandler;
import org.tranql.schema.Association;
import org.tranql.schema.AssociationEnd;
import org.tranql.schema.Attribute;
import org.tranql.schema.Entity;

/**
 *
 *
 * @version $Revision: 323 $ $Date: 2006-02-19 01:29:24 -0500 (Sun, 19 Feb 2006) $
 */
public class PrefetchGroupHandler implements ResultHandler, Serializable {
    private static final ChildAccessor NO_OP_CHILD_ACCESSOR = new NoOpChildAccessor(); 
    
    public static PrefetchGroupHandler buildHandler(FieldTransform[] fields, PrefetchGroup group, IdentityDefinerBuilder idDefBuilder) {
        PrefetchGroupHandler rootLoader = null;

        IdentityHashMap groupToLoader = new IdentityHashMap();
        int attributeIndex = 0;
        PrefetchGroup.PrefetchGroupIterator iterator = group.getIterator();
        while (iterator.hasMorePrefetchGroups()) {
            PrefetchGroup prefetchGroup = iterator.nextPrefetchGroup();
            
            Entity entity = prefetchGroup.getEntity();
            List pkFields = entity.getPrimaryKeyFields();
            FieldTransform[] pkAccessors = new FieldTransform[pkFields.size()];
            System.arraycopy(fields, attributeIndex, pkAccessors, 0, pkAccessors.length);
            IdentityDefiner idDef = idDefBuilder.getIdentityDefiner(entity, pkAccessors);
            
            Collection slotLoaders = new ArrayList();
            for (Iterator iter = pkFields.iterator(); iter.hasNext();) {
                Attribute attribute = (Attribute) iter.next();
                int slotIndex = entity.getAttributes().indexOf(attribute);
                slotLoaders.add(new EmptySlotLoader(slotIndex, fields[attributeIndex++]));
            }
            Attribute[] attributes = prefetchGroup.getAttributes();
            for (int i = 0; i < attributes.length; i++) {
                Attribute attribute = attributes[i];
                if (attribute.isIdentity()) {
                    continue;
                }
                int slotIndex = entity.getAttributes().indexOf(attribute);
                slotLoaders.add(new EmptySlotLoader(slotIndex, fields[attributeIndex++]));
            }
            EmptySlotLoader[] loaders = (EmptySlotLoader[]) slotLoaders.toArray(new EmptySlotLoader[slotLoaders.size()]);
            
            PrefetchGroup parent = prefetchGroup.getParentPrefetchGroup();
            if (null != parent) {
                PrefetchGroup.AssociationEndInfo parentAssociationEndInfo = prefetchGroup.getParentEndTableInfo();
                Entity parentEntity = parent.getEntity();
                AssociationEnd end = parentAssociationEndInfo.getAssociationEnd();
                Association association = end.getAssociation();
                int parentSlot = parentEntity.getAttributes().size() + parentEntity.getAssociationEnds().indexOf(end);
                FieldTransform parentRowAccessor = new FieldAccessor(parentSlot, null);
                ParentAccessor parentAccessor;
                if (end.isOneToOne() || end.isManyToOne()) {
                    parentAccessor = new SingleValueAccessor(parentRowAccessor);
                } else {
                    parentAccessor = new MultiValueAccessor(parentRowAccessor);
                }
                ChildAccessor childAccessor = NO_OP_CHILD_ACCESSOR;
                if (end.isOneToOne() || end.isOneToMany()) {
                    AssociationEnd childEnd = association.getOtherEnd(end);
                    int childSlot = entity.getAttributes().size() + entity.getAssociationEnds().indexOf(childEnd);
                    FieldTransform childRowAccessor = new FieldAccessor(childSlot, null);
                    childAccessor = new ChildAccessorImpl(childRowAccessor);
                }
                PrefetchGroupHandler parentLoader = (PrefetchGroupHandler) groupToLoader.get(parent);
                PrefetchGroupHandler childHandler = parentLoader.addChildLoader(idDef, parentAccessor, childAccessor, loaders);
                groupToLoader.put(prefetchGroup, childHandler);
            } else {
                rootLoader = new PrefetchGroupHandler(idDef, loaders);
                groupToLoader.put(prefetchGroup, rootLoader);
            }
        }
        
        return rootLoader;
    }

    private transient final ResultHandler topHandler;
    private final IdentityDefiner idDefiner;
    private final ParentAccessor parentAccessor;
    private final ChildAccessor childAccessor;
    private final EmptySlotLoader[] slotLoaders;
    private final Collection childHandlers;
    private final ResultHandler delegate;

    public PrefetchGroupHandler(PrefetchGroupHandler prototype, ResultHandler topHandler) {
        this.topHandler = topHandler;
        this.idDefiner = prototype.idDefiner;
        this.parentAccessor = prototype.parentAccessor;
        this.childAccessor = prototype.childAccessor;
        this.slotLoaders = prototype.slotLoaders;
        childHandlers = new ArrayList(prototype.childHandlers);
        this.delegate = new TopHandler(topHandler, idDefiner, slotLoaders, childHandlers);
    }
    
    private PrefetchGroupHandler(IdentityDefiner idDefiner, EmptySlotLoader[] slotLoaders) {
        this.topHandler = null;
        this.idDefiner = idDefiner;
        this.parentAccessor = null;
        this.childAccessor = null;
        this.slotLoaders = slotLoaders;
        delegate = new UndefinedTopHandler();
        childHandlers = new ArrayList();
    }

    private PrefetchGroupHandler(IdentityDefiner idDefiner, ParentAccessor parentAccessor, ChildAccessor childAccessor, EmptySlotLoader[] slotLoaders) {
        this.topHandler = null;
        this.idDefiner = idDefiner;
        this.parentAccessor = parentAccessor;
        this.childAccessor = childAccessor;
        this.slotLoaders = slotLoaders;
        childHandlers = new ArrayList();
        delegate = new ChildHandler(idDefiner, slotLoaders, parentAccessor, childAccessor, childHandlers);
    }

    public Object execute(InTxCache cache, QueryCommand command, Row params, Object arg) throws QueryException {
        CacheContext ctx = new CacheContext(cache, arg);
        command.execute(cache, this, params, ctx);
        return ctx.nestedContext;
    }
    
    public Object wrapContext(InTxCache cache, Object arg) {
        return new CacheContext(cache, arg);
    }

    public Object unwrapContext(Object arg) {
        return ((CacheContext) arg).nestedContext;
    }

    public Object fetched(Row row, Object arg) throws QueryException {
        return delegate.fetched(row, arg);
    }
    
    public Object endFetched(Object arg) throws QueryException {
        CacheContext ctx = (CacheContext) arg;
        ctx.inConstruction.bool = false;
        return arg;
    }

    private PrefetchGroupHandler addChildLoader(IdentityDefiner idDef, ParentAccessor parentAccessor, ChildAccessor childAccessor, EmptySlotLoader[] loaders) {
        PrefetchGroupHandler loader = new PrefetchGroupHandler(idDef, parentAccessor, childAccessor, loaders);
        childHandlers.add(loader);
        return loader;
    }
    
    static class BooleanHolder {
        boolean bool = true;
    }
    
    static class CacheContext {
        CacheRow cacheRow;
        final InTxCache cache;
        Object nestedContext;
        GlobalIdentity lastTopId;
        BooleanHolder inConstruction = new BooleanHolder();
        
        public CacheContext(InTxCache cache, Object nestedContext) {
            this.cache = cache;
            this.nestedContext = nestedContext;
        }
        
        public Object getNestedContext() {
            return nestedContext;
        }
    }
}
