/**
 *
 *  Copyright 2005 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.sql.prefetch;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.tranql.builder.IdentityDefinerBuilder;
import org.tranql.builder.QueryTileBuilder;
import org.tranql.field.FieldAccessor;
import org.tranql.field.FieldTransform;
import org.tranql.ql.AliasedEntity;
import org.tranql.ql.AttributeReference;
import org.tranql.ql.Join;
import org.tranql.ql.Node;
import org.tranql.ql.Query;
import org.tranql.ql.QueryBinding;
import org.tranql.ql.QueryBindingImpl;
import org.tranql.ql.QueryException;
import org.tranql.ql.QuerySource;
import org.tranql.ql.QuerySourceDictionary;
import org.tranql.ql.Select;
import org.tranql.query.CommandFactory;
import org.tranql.query.QueryCloner;
import org.tranql.query.QueryCommand;
import org.tranql.schema.Association;
import org.tranql.schema.AssociationEnd;
import org.tranql.schema.Attribute;
import org.tranql.schema.Entity;
import org.tranql.schema.Schema;
import org.tranql.schema.Association.JoinDefinition;

/**
 *
 *
 * @version $Revision: 323 $ $Date: 2006-02-19 01:29:24 -0500 (Sun, 19 Feb 2006) $
 */
public class PrefetchGroupTransform extends QueryCloner {
    private final CommandFactory commandFactory;
    private final AliasedEntity rootQS;
    private final PrefetchGroup rootGroup;
    private final QuerySourceDictionary qsDictionary;
    private final IdentityDefinerBuilder idDefBuilder;

    public PrefetchGroupTransform(Schema schema, AliasedEntity rootQS, PrefetchGroup rootGroup, QuerySourceDictionary qsDictionary, IdentityDefinerBuilder idDefBuilder) {
        if (rootQS.getEntity() != rootGroup.getEntity()) {
            throw new IllegalArgumentException("QuerySource [" + rootQS + 
                    "] and PrefetchGroup [" + rootGroup + "] do not match.");
        }
        this.commandFactory = schema.getCommandFactory();
        this.rootQS = rootQS;
        this.rootGroup = rootGroup;
        this.qsDictionary = qsDictionary;
        this.idDefBuilder = idDefBuilder;
    }

    private class Context {
        private IdentityHashMap groupToQS = new IdentityHashMap();
        private Node lastSelectClause;
        private Node fromClause;
        private int lastSelectClauseIndex = 0;
        private List resultBindings = new ArrayList();
        private List prefAccessors = new ArrayList();
        private Map selectedAttToIndex = new HashMap();
        private boolean transformed;

        public void setTransformed(boolean transformed) {
            this.transformed = transformed;
        }

        public boolean isTransformed() {
            return transformed;
        }

        public boolean isSubSelect() {
            return null != lastSelectClause;
        }
        
        public void setSelect(Select select) {
            Node selectClause = select.getChild();
            while (null != selectClause) {
                lastSelectClause = selectClause;
                if (selectClause instanceof AttributeReference) {
                    AttributeReference attRef = (AttributeReference) selectClause;
                    selectedAttToIndex.put(attRef.getAttribute().getName(), new Integer(lastSelectClauseIndex));
                }
                lastSelectClauseIndex++;
                selectClause = lastSelectClause.getSibling();
            }
        }

        private void addSelect(AliasedEntity qs, Attribute attribute) {
            if (rootQS.getAlias().equals(qs.getAlias())) {
                Integer index = (Integer) selectedAttToIndex.get(attribute.getName());
                if (null != index) {
                    prefAccessors.add(new FieldAccessor(index.intValue(), attribute.getType()));
                    return;
                }
            }
            AttributeReference reference = new AttributeReference(qs, attribute); 
            lastSelectClause.setSibling(reference);
            lastSelectClause = reference;
            resultBindings.add(new QueryBindingImpl(lastSelectClauseIndex, qs.getEntity(), attribute));
            prefAccessors.add(new FieldAccessor(lastSelectClauseIndex++, attribute.getType()));
        }
        
        public void addFrom(JoinDefinition joinDefinition, QuerySource qs, QuerySource relatedQS) throws QueryException{
            Node condition = QueryTileBuilder.buildJoinCondition(joinDefinition, qs, relatedQS);
            fromClause = new Join(Join.LEFT_JOIN).addChild(fromClause).addChild(qs).addChild(condition);
        }
        
        public QueryBinding[] getPrefetchBindings() {
            return (QueryBinding[]) resultBindings.toArray(new QueryBinding[resultBindings.size()]);
        }
    }

    public QueryCommand transform(QueryCommand command) throws QueryException {
        Context ctx = new Context();
        Query query = (Query) (command.getQuery().visit(this, ctx));

        // query has not been transformed. return it as is.
        if (false == ctx.isTransformed()) {
            QueryCommand transformedCommand = commandFactory.createQuery(query);
            return transformedCommand;
        }

        // create a new query taking into account the added prefetch bindings.
        Node child = query.getChild();
        QueryBinding[] prefetchBindings = ctx.getPrefetchBindings();
        QueryBinding[] sourceTransforms = query.getResultTransforms();
        QueryBinding[] resultTransforms = new QueryBinding[sourceTransforms.length + prefetchBindings.length];
        System.arraycopy(sourceTransforms, 0, resultTransforms, 0, sourceTransforms.length);
        System.arraycopy(prefetchBindings, 0, resultTransforms, sourceTransforms.length, prefetchBindings.length);
        Query newQuery = new Query(query.getParamTransforms(), resultTransforms);
        newQuery.setResultAccessors(query.getResultAccessors());
        newQuery.addChild(child);
        QueryCommand transformedCommand = commandFactory.createQuery(newQuery);

        // create the wrapping prefetch query command.
        FieldTransform[] prefetchAccessors = (FieldTransform[]) ctx.prefAccessors.toArray(new FieldTransform[ctx.prefAccessors.size()]);
        PrefetchGroupHandler handler = PrefetchGroupHandler.buildHandler(prefetchAccessors, rootGroup, idDefBuilder);
        return new PrefetchQueryCommand(transformedCommand, handler);
    }

    public Query transform(Query query) throws QueryException {
        return (Query) query.visit(this, new Context());
    }

    public Object visit(Query node, Object param) throws QueryException {
        Query to = (Query) super.visit(node, param);
        
        Context ctx = (Context) param;
        if (false == ctx.isTransformed()) {
            return to;
        }

        return to;
    }

    public Object visit(Select node, Object param) throws QueryException {
        Select to = (Select) super.visit(node, param);

        Context ctx = (Context) param;
        if (ctx.isSubSelect()) {
            return to;
        }
        ctx.setSelect(to);
        if (to.isDistinct()) {
            ctx.setTransformed(false);
            return to;
        }
        
        ctx.setTransformed(true);
        boolean first = true;
        PrefetchGroup.PrefetchGroupIterator iterator = rootGroup.getIterator();
        while (iterator.hasMorePrefetchGroups()) {
            PrefetchGroup group = iterator.nextPrefetchGroup();
            Entity entity = group.getEntity();
            AliasedEntity qs;
            if (first) {
                first = false;
                qs = rootQS;
            } else {
                qs = new QuerySource(entity, qsDictionary.getNextAlias());
            }
            ctx.groupToQS.put(group, qs);
            
            List pkFields = entity.getPrimaryKeyFields();
            for (Iterator iter = pkFields.iterator(); iter.hasNext();) {
                Attribute attribute = (Attribute) iter.next();
                ctx.addSelect(qs, attribute);
            }
            Attribute[] attributes = group.getAttributes();
            for (int i = 0; i < attributes.length; i++) {
                Attribute attribute = attributes[i];
                if (attribute.isIdentity()) {
                    continue;
                }
                ctx.addSelect(qs, attribute);
            }
        }
        return to;
    }

    public Object visit(QuerySource node, Object param) throws QueryException {
        Context ctx = (Context) param;
        QuerySource to = (QuerySource) super.visit(node, param);
        if (false == rootQS.getAlias().equals(to.getAlias())) {
            return to;
        }
        
        ctx.fromClause = to;
        
        PrefetchGroup.PrefetchGroupIterator iterator = rootGroup.getIterator();
        while (iterator.hasMorePrefetchGroups()) {
            PrefetchGroup group = iterator.nextPrefetchGroup();
            QuerySource qs = (QuerySource) ctx.groupToQS.get(group);
            
            PrefetchGroup.AssociationEndInfo[] endInfo = group.getEndTableInfo();
            for (int i = 0; i < endInfo.length; i++) {
                PrefetchGroup.AssociationEndInfo info = endInfo[i];
                AssociationEnd end = info.getAssociationEnd();
                Association association = end.getAssociation();
                JoinDefinition joinDefinition;
                if (end.isManyToMany()) {
                    Entity mtmEntity = association.getManyToManyEntity();
                    QuerySource mtmQS = new QuerySource(mtmEntity, qsDictionary.getNextAlias());
                    JoinDefinition mtmJoinDefinition = association.getLeftJoinDefinition();
                    joinDefinition = association.getRightJoinDefinition();
                    if (mtmJoinDefinition.getPKEntity() != qs.getEntity()) {
                        joinDefinition = mtmJoinDefinition;
                        mtmJoinDefinition = association.getRightJoinDefinition();
                        if (mtmJoinDefinition.getPKEntity() != qs.getEntity()) {
                            throw new AssertionError();
                        }
                    }
                    ctx.addFrom(mtmJoinDefinition, mtmQS, qs);
                    
                    qs = mtmQS;
                } else {
                    joinDefinition = association.getJoinDefinition();
                }
                QuerySource relatedQS = (QuerySource) ctx.groupToQS.get(info.getPrefetchGroup());
                ctx.addFrom(joinDefinition, relatedQS, qs);
            }
        }        
        return ctx.fromClause;
    }
}
